/*
 *
 * Copyright (C) Mellanox Technologies Ltd. 2001-2014.  ALL RIGHTS RESERVED.
 *
 * Based on:
 *	pca954x.c from Rodolfo Giometti <giometti@linux.it>
 *   Copyright (c) 2008-2009 Eurotech S.p.A. <info@eurotech.it>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 */

#include <linux/module.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/device.h>
#include <linux/i2c.h>
#include <linux/i2c-mux.h>
#include <asm/io.h>
#include <linux/platform_device.h>

#include "cpld_mux.h"

#define CPLD_MUX_MOD_DESCRIPTION    "I2C CPLD Mux driver. Build:" " "__DATE__" "__TIME__
#define CPLD_MUX_MOD_VERSION        "1.0.0"

static int dbg_lvl = CPLD_MUX_DFLT_DBG_LVL;
module_param(dbg_lvl, int, 0644);

struct cpld_mux_common_data* cpld_mux_comm_data = NULL;

static const struct mux_desc muxes[] = {
	[cpld_mux_tor] = {
		.nchans = CPLD_TOR_MUX_MAX_NCHANS,
		.muxtype = lpc_access,
	},
	[cpld_mux_mgmt] = {
		.nchans = CPLD_MGMT_MUX_MAX_NCHANS,
		.muxtype = lpc_access,
	},
	[cpld_mux_module] = {
		.nchans = CPLD_TOR_MUX_MAX_NCHANS,
		.muxtype = i2c_access,
	},
};

static const struct i2c_device_id cpld_mux_id[] = {
	{ "cpld_mux_tor", cpld_mux_tor },
	{ "cpld_mux_mgmt", cpld_mux_mgmt },
	{ "cpld_mux_module", cpld_mux_module },
	{ }
};
MODULE_DEVICE_TABLE(i2c, cpld_mux_id);

/* Write to mux register. Don't use i2c_transfer()/i2c_smbus_xfer()
   for this as they will try to lock adapter a second time */
static int cpld_mux_reg_write(struct i2c_adapter *adap,
			     struct i2c_client *client, u8 val, enum mux_type muxtype)
{
	int ret = -ENODEV;
    struct cpld_mux_platform_data *pdata = dev_get_platdata(&client->dev);

    if (muxtype == lpc_access) {
        outb(val, pdata->addr); // Addr = CPLD base + offset
        CPLD_MUX_LOG_DBG(4, "LPC Write reg 0x%x val %d\n", pdata->addr, val);
        ret = 1;
    }
    else {
        if (adap->algo->master_xfer) {
            struct i2c_msg msg;
            u8 msgbuf[] = {pdata->sel_reg_addr, val};

            msg.addr = pdata->addr;
            msg.flags = 0;
            msg.len = 2;
            msg.buf = msgbuf;
            ret = adap->algo->master_xfer(adap, &msg, 1);
        } else {
            dev_err(&client->dev,
                    "SMBus isn't supported on this adapter");
        }
    }

	return ret;
}

static int cpld_mux_select_chan(struct i2c_adapter *adap,
			       void *client, u32 chan)
{
	struct cpld_mux *data = i2c_get_clientdata(client);
	const struct mux_desc *mux = &muxes[data->type];
	u8 regval;
	int ret = 0;
    struct i2c_client* cl = (struct i2c_client*)client;
    struct cpld_mux_platform_data *pdata = dev_get_platdata(&cl->dev);

	regval = pdata->first_channel + chan;

    CPLD_MUX_LOG_DBG(3, "%s client %s, select channel %d\n", cl->dev.kobj.name, cl->name, regval);
	/* Only select the channel if its different from the last channel */
	if (data->last_chan != regval) {
		ret = cpld_mux_reg_write(adap, client, regval, mux->muxtype);
		data->last_chan = regval;
	}

	return ret;
}

static int cpld_mux_deselect_mux(struct i2c_adapter *adap,
				void *client, u32 chan)
{
	struct cpld_mux *data = i2c_get_clientdata(client);
    const struct mux_desc *mux = &muxes[data->type];

    if (dbg_lvl >= 2){
        struct i2c_client* cl = (struct i2c_client*)client;
        struct cpld_mux_platform_data *pdata = dev_get_platdata(&cl->dev);
        chan += pdata->first_channel;
        CPLD_MUX_LOG_DBG(3, "%s client %s, deselect channel %d\n", cl->dev.kobj.name, cl->name, chan);
    }

	/* Deselect active channel */
	data->last_chan = 0;
	return cpld_mux_reg_write(adap, client, data->last_chan, mux->muxtype);
}

ssize_t show_channel(struct device *dev, struct device_attribute *attr, char *buf)
{
    ssize_t rc;
    struct i2c_client *client;
    struct cpld_mux_platform_data* pdata;
    struct cpld_mux *data;

    pdata = dev_get_platdata(dev);
    client = container_of(dev, struct i2c_client, dev);
    data = i2c_get_clientdata(client);
    rc = sprintf(buf, "CPLD_MUX 0x%x last selected channel: %d\n", \
                     pdata->id, data->last_chan);

    return rc+1;
}

ssize_t store_channel(struct device *dev, struct device_attribute *attr, const char *buf, size_t cnt)
{
    int rc;
    u32 chan, prev_chan;
    struct i2c_client *client;
    struct i2c_adapter *adap;
    struct cpld_mux_platform_data* pdata;
    struct cpld_mux *data;

    if (!buf || cnt == 1) {
        CPLD_MUX_LOG_ERROR("Incorrect empty str input\n");
        return cnt;
    }

    pdata = dev_get_platdata(dev);
    client = container_of(dev, struct i2c_client, dev);
    data = i2c_get_clientdata(client);
    adap = to_i2c_adapter(&(client->dev));

    prev_chan = data->last_chan;
    chan = (u32)simple_strtol(buf, NULL, 10);
    rc = cpld_mux_select_chan(adap, client, chan);
    if (rc < 0) {
        CPLD_MUX_LOG_DBG(1, "CPLD_MUX 0x%x change channel %d failed\n", \
                         pdata->id, chan);
    } 
    else {
        CPLD_MUX_LOG_DBG(1, "CPLD_MUX 0x%x last selected channel changed from %d to %d\n", \
                         pdata->id, prev_chan, chan);
    }

    return cnt;
}

static DEVICE_ATTR(channel, 0644, show_channel, store_channel);

static struct attribute *cpld_mux_attrs[] = {
    &dev_attr_channel.attr,
	NULL
};

static struct attribute_group cpld_mux_attr_grp = {
    .attrs = cpld_mux_attrs
};
 
int cpld_mux_sysfs_device_create(struct device* dev)
{
    struct cpld_mux_platform_data* pdata = dev_get_platdata(dev);

	if (sysfs_create_group(&(dev->kobj), &cpld_mux_attr_grp)) {
        CPLD_MUX_LOG_ERROR("Failed to create sysfs cpld_mux group\n");
        pdata->mux_kobj = NULL;
		return -EACCES;
    }
    pdata->mux_kobj = &(dev->kobj);

    return 0;
}

void cpld_mux_sysfs_device_delete(struct cpld_mux_platform_data *pdata)
{
    if (pdata->mux_kobj) 
        sysfs_remove_group(pdata->mux_kobj, &cpld_mux_attr_grp);

    pdata->mux_kobj = NULL;
} 

int cpld_mux_reset(struct i2c_client *client)
{
	struct i2c_adapter *adap = to_i2c_adapter(client->dev.parent);
    struct cpld_mux *data = i2c_get_clientdata(client);

	if (!i2c_check_functionality(adap, I2C_FUNC_SMBUS_BYTE))
		return -ENODEV;

	data->last_chan = 0;		   /* force the first selection */

	return 0;
}
EXPORT_SYMBOL_GPL(cpld_mux_reset);

int cpld_mux_block_unblock(struct i2c_client *client, u8 blk)
{
	struct cpld_mux *data = i2c_get_clientdata(client);
	int i;
	for (i = 0; i < CPLD_MGMT_MUX_MAX_NCHANS; ++i) {
		if (data && data->virt_adaps[i])
			data->virt_adaps[i]->blocked = blk;
	}
	return 0;
}
EXPORT_SYMBOL_GPL(cpld_mux_block_unblock);

/*
 * I2C init/probing/exit functions
 */
static int cpld_mux_probe(struct i2c_client *client,
			 const struct i2c_device_id *id)
{
	struct i2c_adapter *adap = to_i2c_adapter(client->dev.parent);
	struct cpld_mux_platform_data *pdata = dev_get_platdata(&client->dev);
	int num, force;
	struct cpld_mux *data;
	int ret = -ENODEV;

    CPLD_MUX_LOG_DBG(2, "%s client %s probe\n", client->dev.kobj.name, client->name);
	if (!i2c_check_functionality(adap, I2C_FUNC_SMBUS_BYTE))
		goto err;

	data = kzalloc(sizeof(struct cpld_mux), GFP_KERNEL);
	if (!data) {
		ret = -ENOMEM;
		goto err;
	}

    i2c_set_clientdata(client, data);

	data->type = id->driver_data;
	data->last_chan = 0;		   /* force the first selection */

	/* Now create an adapter for each channel */
	for (num = 0; num < muxes[data->type].nchans; num++) {
		force = 0;			  /* dynamic adap number */
		// class = 0;			  /* no class by default */
		if (pdata) {
			if (num < pdata->num_modes) {
				/* force static number */
				force = pdata->modes[num].adap_id;
				// class = pdata->modes[num].class;
			} else
				/* discard unconfigured channels */
				break;
		}

		data->virt_adaps[num] =
			i2c_add_mux_adapter(adap, &client->dev, client,
				force, num, 0, cpld_mux_select_chan,
				(pdata && pdata->modes[num].deselect_on_exit)
					? cpld_mux_deselect_mux : NULL);

		if (data->virt_adaps[num] == NULL) {
			ret = -ENODEV;
			dev_err(&client->dev,
				"failed to register multiplexed adapter"
				" %d as bus %d\n", num, force);
			goto virt_reg_failed;
		}
        else {
            CPLD_MUX_LOG_DBG(2, "Added i2c addapter, num %d, deselect %s, force %d, name %s, parent name %s\n", \
                             num, (pdata && pdata->modes[num].deselect_on_exit) ? "yes" : "no", \
                             force, data->virt_adaps[num]->name, adap->name);
        }
	}
    if (cpld_mux_sysfs_device_create(&(client->dev))) {
        ret = -ENODEV;
        dev_err(&client->dev, "failed create sysfs for cpld_mux_%x\n", pdata->id);
        goto virt_reg_failed;
    }
    else {
        list_add(&(pdata->dev_list), &(cpld_mux_comm_data->head_list));
    }

	CPLD_MUX_LOG_DBG(1, "Registered %d multiplexed busses for CPLD I2C mux %s, id 0x%x\n", num, client->name, pdata->id);

	return 0;

virt_reg_failed:
	for (num--; num >= 0; num--)
		i2c_del_mux_adapter(data->virt_adaps[num]);
	kfree(data);
err:
    kfree(pdata);
	return ret;
}

static int cpld_mux_remove(struct i2c_client *client)
{
	struct cpld_mux *data = i2c_get_clientdata(client);
	const struct mux_desc *mux = &muxes[data->type];
    struct cpld_mux_platform_data *pdata = dev_get_platdata(&client->dev);
	int i;

    CPLD_MUX_LOG_DBG(2, "%s client %s remove\n", client->dev.kobj.name, client->name);
	for (i = 0; i < mux->nchans; ++i)
		if (data->virt_adaps[i]) {
		    CPLD_MUX_LOG_DBG(2, "%s chan %d, adapter %s\n", __FUNCTION__, i, data->virt_adaps[i]->name);
			i2c_del_mux_adapter(data->virt_adaps[i]);
			data->virt_adaps[i] = NULL;
		}
    cpld_mux_sysfs_device_delete(pdata);
    list_del(&(pdata->dev_list));
    kfree(pdata);

	kfree(data);
	return 0;
}

static struct i2c_driver cpld_mux_driver = {
	.driver		= {
		.name	= "CPLD_MUX",
		.owner	= THIS_MODULE,
	},
	.probe		= cpld_mux_probe,
	.remove		= cpld_mux_remove,
	.id_table	= cpld_mux_id,
};

static int __init cpld_mux_init(void)
{
    int rc;

    cpld_mux_comm_data = kzalloc(sizeof(struct cpld_mux_common_data), GFP_KERNEL);
    if (!cpld_mux_comm_data) {
        CPLD_MUX_LOG_ERROR("Alloc cpld_mux_common_data failed\n");
        return -ENOMEM;
    }

    INIT_LIST_HEAD(&(cpld_mux_comm_data->head_list));

	rc = i2c_add_driver(&cpld_mux_driver);
    if (rc) {
        CPLD_MUX_LOG_ERROR("Add %s i2c driver failed (%d)\n", CPLD_MUX_DEVICE_NAME, rc);
        kfree(cpld_mux_comm_data);
    }

    return rc;
}

static void __exit cpld_mux_exit(void)
{
    struct cpld_mux_platform_data* curr, *next;

    list_for_each_entry_safe(curr, next, &(cpld_mux_comm_data->head_list), dev_list) {
        cpld_mux_sysfs_device_delete(curr);
    }

    kfree(cpld_mux_comm_data);
	i2c_del_driver(&cpld_mux_driver);
}

module_init(cpld_mux_init);
module_exit(cpld_mux_exit);

MODULE_AUTHOR("Mellanox Technologies (Michael Shych)");
MODULE_DESCRIPTION(CPLD_MUX_MOD_DESCRIPTION);
MODULE_VERSION(CPLD_MUX_MOD_VERSION);
MODULE_LICENSE("GPL v2");

MODULE_PARM_DESC(dbg_lvl, " debug level, 0-3, 0 - disabled, 1 - default");




