/**
 *
 * Copyright (C) 2012-2014, Mellanox Technologies Ltd.  ALL RIGHTS RESERVED.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 */

#ifndef __SWITCHX_H_
#define __SWITCHX_H_

/*  Command Interface Through i2c
    All commands can be submitted through PCIe bus, while only certain commands can be submitted
    through the i2c bus. The i2c addressing is aligned with the PCIe addressing, and the location
    of the Secondary CIR is as shown in Section 249, “SwitchX BAR0 Layout,” on page 288.
    Some of the commands use mailboxes. In order to use mailboxes through the i2c, special area is
    reserved on the i2c address space that can be used for input and output mailboxes. 
    Such mailboxes are called Local Mailboxes. When using a local mailbox, software should specify 0
    as the Input/Output parameters.
    The location of the Local Mailbox addresses on the i2c space can be retrieved through the
    QUERY_FW command.
    SwitchX can be managed through i2c, PCIe, SGMII Management Ports or in-band.
    Only basic initialization commands supported.

    Command interface comprises a set of registers through which software can submit commands to
    the device. Software constructs the command. The command is written into the command interface
    register. Software can then poll on the command interface register until completed or wait for
    interrupt. To pass data into the command and to receive data back from the hardware, software can
    use either the command register inline data, or to include a pointer to mailboxes that reside in system
    memory.

    In order to address a PCI device it must be mapped into the I/O port address space or the memory
    mapped address space of the system. The switch device drivers will program the Base Address Registers
    to inform the device of its address mapping by writing configuration commands to the PCI controller.
    The SwitchX driver accesses the device through PCIe Base Address Register 0 (BAR0).
    (./arch/powerpc/include/asm/pci-bridge.h: field cfg_addr of struct pci_controller).
    There are two command interface registers, primary and secondary
    Primary located within Switch BAR0 at addresses 71000h-71018h, for command submission through PCIe.
    Secondary located within Switch BAR0 at addresses 72000h-72018h, for command submission through i2c. 
    The secondary CIR supports restricted set of commands, only:
     - QUERY_FW        Query Firmware 
     - QUERY_BOARDINFO Query Board Information
     - CONFIG_PROFILE  Configure Switch Profile
     - ACCESS_REG      Access an EMAD supported register

    Prior to submitting a command software must ensure that CIR is ready to receive the
    new command (see further discussion). To submit a command, the software should fill the CIR and
    set the go bit. CIR supports two types of commands: immediate and event. Immediate commands,
    complete as soon as the go bit is cleared by the hardware. Event commands completion is reported
    to an event queue. The CIR is ready to receive a new command when the go bit is clear.

    Full list of SwitchX commands
    Opcode Command         Description                          PICe I2C Reference
                    Initialization and Teardown Commands
    0x4    QUERY_FW        Query Firmware                         v   v  Section 8.5.1.1
    0x6    QUERY_BOARDINFO Query Board Information                v   v  Section 8.5.1.2
    0x3    QUERY_AQ_CAP    Query Asynchronous Queues Capabilities v      Section 8.5.1.3
    0xFFF  MAP_FA          Map Firmware Area v Section 8.5.1.4
    0xFFE  UNMAP_FA        Unmap Firmware Area                    v     Section 8.5.1.5
    0x100  CONFIG_PROFILE  Configure Switch Profile               v   v Section 8.5.1.6
                    General Commands
    0x40   ACCESS_REG      Access an EMAD supported register      v   v Section 8.5.2.1
                    Asynchronous Queues Commands
    0x201  SW2HW_DQ        Software to Hardware DQ                v     Section 8.5.4.1
    0x202  HW2SW_DQ        Hardware to Software DQ                v     Section 8.5.4.2
    0x1E   2ERR_DQ         To Error DQ                            v     Section 8.5.4.3
    0x22   QUERY_DQ        Query DQ                               v     Section 8.5.4.4
    0x16   SW2HW_CQ        Software to Hardware CQ                v     Section 8.5.4.5
    0x17   HW2SW_CQ        Hardware to Software CQ                v     Section 8.5.4.6
    0x18   QUERY_CQ        Query CQ                               v     Section 8.5.4.7
    0x13   SW2HW_EQ        Software to Hardware EQ                v     Section 8.5.4.8
    0x14   HW2SW_EQ        Hardware to Software EQ                v     Section 8.5.4.9
    0x15   QUERY_EQ        Query EQ                               v     Section 8.5.4.10
                    InfiniBand Commands
    0x203  INIT_MAD_DEMUX Initialize MAD De-multiplexing          v     Section 8.5.5.1
    0x24   MAD_IFC        Management Datagram Interface           v     Section 8.5.5.2
*/

#define VSD_VENDOR_ID       0x15B3
#define MAX_EMAD_FRAME_SIZE 1518
#define DIR_ROUTE_TLV_SIZE   132
#define END_TLV_SIZE           4

/* QoS Registers */
#define SX_QCAP_REG  0x4001 // QoS Capabilities Register
#define SX_QPTL_REG  0x4002 // QoS Priority Trust Level Register 
#define SX_QDPM_REG  0x4003 // QoS DSCP to Priority Mapping Register
#define SX_QPBC_REG  0x4004 // QoS Policer Binding and Configuration Register
#define SX_QETCR_REG 0x4005 // QoS ETS TC Configuration Register
#define SX_QEGCS_REG 0x4006 // QoS ETS General Configuration Register
#define SX_QPDP_REG  0x4007 // QoS Port Default Priority
#define SX_QPRT_REG  0x4008 // QoS Priority Regeneration Table
#define SX_QSPTC_REG 0x4009 // QoS Stacking Port TClass
#define SX_QTCT_REG  0x400A // QoS Switch Traffic Class Table
#define SX_QSTCT_REG 0x400B // Switch Stacking TClass Table

/* Port Registers */
#define SX_PCAP_REG  0x5001 // Port Switch Partition Allocation (Configure Port Membership in Switch Partitions)
#define SX_PMLP_REG  0x5002 // Ports Module to Local Port Register 
#define SX_PMTU_REG  0x5003 // Port MTU Register
#define SX_PTYS_REG  0x5004 // Port Type and Speed Register 
#define SX_PPAD_REG  0x5005
#define SX_PAOS_REG  0x5006 // Ports Administrative and Operatoinal Status Register
#define SX_PFCC_REG  0x5007
#define SX_PPCNT_REG 0x5008
#define SX_PUDE_REG  0x5009
#define SX_PLIB_REG  0x500A // Port Local Port to IB Port Mapping Register
#define SX_PPTB_REG  0x500B
#define SX_PBMC_REG  0x500C
#define SX_PSPA_REG  0x500D
#define SX_PELC_REG  0x500E // Port Extended Link Capabilities Register
#define SX_PVLC_REG  0x500F // Port Virtual Lane Capabilities Register

#define SX_SPZR_REG  0x6002 // Swicth Port Zero Infiniband Register

/* Host Interface Control Registers */
#define SX_HCAP_REG  0x7001 // Host Interface Capabilities
#define SX_HTGT_REG  0x7002 // Host Trap Group Table
#define SX_HPKT_REG  0x7003 // Host Packet Trap
#define SX_HDRT_REG  0x7004 // Host Direct Route Table Register
#define SX_HESPR_REG 0x7005 // Host EMAD Stacking_EN Path Response Register

/* Miscellaneous Control Registers */
#define SX_MFCR_REG  0x9001 // Misc Fan Control Register
#define SX_MFSC_REG  0x9002 // Misc Fan Speed Control Register
#define SX_MFSM_REG  0x9003 // Misc Fan Speed Measurements Register
#define SX_MFSL_REG  0x9004 // Misc Fan Speed Limit Register
#define SX_FORE_REG  0x9007 // Fan Out of Range Event Register
#define SX_MTCAP_REG 0x9009 // Misc Temperature Capabilities Register
#define SX_MTMP_REG  0x900A // Misc Temperature Register
#define SX_MTWE_REG  0x900B // Misc Temperature Warning Event Register
#define SX_MFPA_REG  0x9010 // Misc Flash Params Register
#define SX_MFBA_REG  0x9011 // Misc Flash Block Access Register
#define SX_MFBE_REG  0x9012 // Misc Flash Block Erase
#define SX_MCIA_REG  0x9014 // Misc Cable Info Access Register
#define SX_MCAS_REG  0x9015 // Misc Cable Stamping Register
#define SX_MMIA_REG  0x9016 // Misc Manufacturing Information Access Register
#define SX_MMDIO_REG 0x9017 // Misc MDIO Interface Register
#define SX_MMEM_REG  0x9018 // Misc Memory Register
#define SX_MPGR_REG  0x9019 // Misc Power Governor Register Fields
#define SX_MPAT_REG  0x901A // Misc Port Analyzer Table Register
#define SX_MPAR_REG  0x901B // Misc Port Analyzer Register
#define SX_MFM_REG   0x901D // Misc Fabric Memory  Register
#define SX_MHSR_REG  0x901E // Misc Health State Register
#define SX_MJTAG_REG 0x901F // Misc JTAG Register
#define SX_MGIR_REG  0x9020 // Misc General Information Register
#define SX_MPFM_REG  0x9022 // Misc PCIe Failure Mode Register
#define SX_MSCI_REG  0x902A // Misc System CPLD Information Register

/* Command Interface Register 
   Read and write accesses to the CIR must be DWORD-sized and DWORD aligned 
   (i.e., the access address and access size must both be multiplications of 4 bytes)
*/
#define  CIR_GO_SW           0x0 // Software ownership
#define  CIR_GO_HW           0x1 // Hardware ownership
#define  CIR_EVENT_NO_REPORT 0x0
#define  CIR_EVENT_REPORT    0x1
#define  CIR_STATUS_OFF      0x18
#define  IN_PARAM_OFFSET     0x00
#define  IN_MODIFIER_OFFSET  0x08
#define  OUT_PARAM_OFFSET    0x0c
#define  TOKEN_OFFSET        0x14
#define  STATUS_OFFSET       0x18
#define  OPMOD_SHIFT         12
#define  EVENT_BIT           22
#define  GO_BIT              23
struct reg_cir_layout {
        u32 in_param_h;        // Input Parameter: parameter or pointer to input mailbox 
        u32 in_param_l;        // Input Parameter: parameter or pointer to input mailbox 
        u32 input_modifier;    // Input Parameter Modifier
        u32 out_param_h;       // Output Parameter: parameter or pointer to output mailbox 
        u32 out_param_l;       // Output Parameter: parameter or pointer to output mailbox 
#ifdef __BIG_ENDIAN
        u16 token;             // Software assigned token to the command, to uniquely identify it. 
                               // The token is returned to the software in the EQE reported.
        u16 reserved1;         // 
        u8 status;             // Command execution status report. Valid only if command interface
                               // in under SW ownership (Go bit is cleared)
                               // 0 - command completed without error. If different than zero,
                               // command execution completed with error. Syndrome encoding is
                               // depended on command executed and is defined for each command
        u8 go:1;                // Go (0=Software ownership for the CIR, 1=Hardware ownership for the CIR)
                                // Software can write to the CIR only if go bit is cleared.
                                // Software must set the go bit to trigger HW to execute the command.
                                // Software must not write to the go bit a value other than 1.
        u8 event:1;             // Event Request
                                // 0 - Don’t report event (software will poll the GO bit)
                                // 1 - Report event to EQ when the command completes
        u8 reserved2:6;         // Event Request
        u16 opcode_modifier:4;  // Opcode Modifier, see specific description for each command.
        u16 opcode:12;          // Command opcode
#else
#if 0
        u16 reserved1;
        u16 token;
        u16 opcode:12;
        u16 opcode_modifier:4;
        u8 reserved2:6;
        u8 event:1;
        u8 go:1;
        //u8 go:1;
        //u8 event:1;
        //u8 reserved2:6;
#endif
        u16 reserved1;
        u16 token;
        u8 status;
        u8 reserved2:6;
        u8 event:1;
        u8 go:1;
        u16 opcode:12;
        u16 opcode_modifier:4;
#endif
};

/* Software Reset Register */
#define SW_RESET 0x1
struct reg_sw_reset_layout {
#ifdef __BIG_ENDIAN
        u32 reserved1:31;       //
        u32 sw_reset:1;         // bit 0 Software Reset. Write 1 to reset the device.
#else
        u32 sw_reset:1;
        u32 reserved1:31;
#endif
};

/* DoorBell Registers 
   Located on the device BAR. The location of the doorbell registers can
   be queried through the QUERY_FW
*/
struct reg_dorbell_layout {
        u8 reserved[100];       // ??? Dorbell size 
};

/* Message Signalled Interrupt Register */
struct reg_msix_layout {
        u8 reserved[100];       // ??? Message Signalled Interrupt Register size
};

/* Clear Interrupt Register */
struct reg_clr_irq_layout {
        u32 clr_int;            // Clear Interrupt. Write transactions to this register will clear
                                // (de-assert) the virtual interrupt output pins. The value
                                // to be written in this register is obtained through the
                                // QUERY_BOARDINFO command.
                                // This register is write-only. Reading from this register will cause
                                // undefined result
};

/* SwitchX BAR0 Layout
   - Command Interface Register (CIR) is used to submit commands mainly for
     initial configuration of the device. Two commands registers are supported
     - Primary CIR - for command submission through PCIe
     - Secondary CIR - for command submission through i2c. The secondary CIR supports 
       restricted set of commands.
   - Doorbell registers are used to update the HW with new descriptors posted to send/receive descriptor
     queues or management of completion and event queues.
   - Clear Interrupt Register
   - MSIX Registers - SwitchX can generate interrupts in the form of PCI interrupt or Message Signalled Interrupt (MSIX).
   - SW Reset Register
*/
struct sx_bar0_layout {
#define SX_BAR0_OFFSET_PRIM    0x71000
        struct reg_cir_layout prim_cmd_if;   // 71000h-71018h Primary Command Interface Register
#define SX_BAR0_OFFSET_SECOND  0x72000
        struct reg_cir_layout second_cmd_if; // 72000h-72018h Secondary Command Interface Register
#define SX_SW_RESET_OFFSET     0xF0010
        struct reg_sw_reset_layout sw_reset; // F00010h SW Reset
        struct reg_dorbell_layout door_bell; // Doorbell Registers
        struct reg_clr_irq_layout irq;       // Clear Interrupt Register
        struct reg_msix_layout msix;         // MSIX Register
};

#define SX_MAX_REG_SIZE 128

/* PSPA - Port Switch Partition Allocation Register Layout */
struct reg_pspa_layout {
#define SWID_PART_STACKING_PORT 254
#define SWID_PART_DISABLED_PORT 255
#ifdef __BIG_ENDIAN
        u8 swid;               // Switch Partition ID to associate {port, sub_port} with
                               // Switch Partitions are numbered from 0 to 7 inclusively.
                               // Switch Partition 254 indicates stacking port.
                               // Switch Partition 255 indicates a disabled port.
        u8 local_port;         // Local port number
        u8 sub_port;           // Virtual port within the physical port. 0 when virtual ports are not enabled on the port.
        u8 reserved1;          //
#else
        u8 reserved1;
        u8 sub_port;
        u8 local_port;
        u8 swid;
#endif
        u32 reserved2;
};

/* PELC - Port Extended Link Capabilities Register Layout */
#define MLP_CAP_LWZ  0x20 // Bit 5: Link Width Control (LWZ)
#define MLP_CAP_EEE  0x10 // Bit 4: Energy Efficient Ethernet (EEE)
#define MLP_CAP_LLR  0x08 // Bit 3: Link Level Rery (LLR)
#define MLP_CAP_CBF  0x04 // Bit 2: Credit-Based Flow Control (CBF)
#define MLP_CAP_FEC  0x02 // Bit 1: Forward Error Correction (FEC)
#define MLP_CAP_AUTO 0x01 // Bit 0: MLPN Autonegotiation
struct reg_pelc_layout {
#ifdef __BIG_ENDIAN
        u8 reserved0;
        u8 local_port;         // Local port number
        u16 reserved1;
#else
        u16 reserved1;
        u8 local_port;
        u8 reserved0;
#endif
        u32 reserved2;
        u64 mlp_capability;    // Port Extended Property supported (bitmask). RO
        u64 mlp_admin;         // Link administratively enabled capabilities by software. 
                               // Same encoding as mlp_capability fields.
                               // HW furthers negotiates the enabled capabilities with remote end of the link. RW
        u64 mlp_active;        // Link operational mode. Same encoding as mlp_capability fields. RO
        u64 mlp_peer;          // Link peer capabilities. Same encoding as mlp_capability fields.
                               // This should be the result of the MLPN advertisement phase. RO
};

/* QSPTC - QoS Stacking Port TClass Table */
struct reg_qsptc_layout {
#ifdef __BIG_ENDIAN
        u8 local_iport;        // Local ingress port number (must be stacking port)
        u8 local_eport;        // Local egress port number (must be stacking port)
        u8 reserved1:5;        // 
        u8 itclass:3;          // Received tclass
        u8 reserved2;
        u32 reserved3:29;      // 
        u32 tclass:3;          // Regenerated stacking traffic class for received packet on stacking port with traffic itclass
                               // The default value for the Stacking Port TClass Regeneration Table is tclass[itclass=i]=i
#else
        u8 reserved2;
        u8 itclass:3;
        u8 reserved1:5;
        u8 local_eport;
        u8 local_iport;
        u32 tclass:3;
        u32 reserved3:29;
#endif
};

/* QSTCT - QoS Switch Stacking TClass Table */
struct reg_qstct_layout {
#ifdef __BIG_ENDIAN
        u8 swid;               // SWitch partition ID
        u8 reserved1;          // 
        u8 reserved2:5;        //   
        u8 prio:3;             // Ingress Priority
        u8 reserved3;
        u8 reserved4[3];
        u8 reserved5:5;        // 
        u8 utclass:3;          // Stacking traffic class used for unicast packets with priority prio
                               // Default values are: utclass[prio=0]=1, utclass[prio=1]=0, utclass[prio=<i>]=i (for i>1)
        u8 reserved6[3];
        u8 reserved7:5;        // 
        u8 mtclass:3;          // Stacking traffic class used for multicast packets with priority prio
                               // Default values are: mtclass[prio=0]=1, mtclass[prio=1]=0, mtclass[prio=<i>]=i (for i>1)
#else
        u8 reserved3;
        u8 prio:3;
        u8 reserved2:5;
        u8 reserved1;
        u8 swid;
        u8 utclass:3;
        u8 reserved5:5;
        u8 reserved4[3];
        u8 mtclass:3;
        u8 reserved7:5;
        u8 reserved6[3];
#endif
};

/* PMTU - Port MTU Register */
struct reg_pmtu_layout {
#ifdef __BIG_ENDIAN
        u8 reserved1;          // 
        u8 local_port;         // Local port number / index
        u16 reserved2;         // 
        u16 max_mtu;           // Maximum MTU supported on the port (Read Only)
                               // MTU depends on the port type. When port type (Eth/IB/FC) is configured, the relevant MTU
                               // is reported. When the port type is not configured, the minimum is reported.
        u16 admin_mtu;         // Administratively configured MTU on the port. Must be smaller or equal to max_mtu RW 
        u16 reserved3;         // 
        u16 oper_mtu;          // Operational MTU. This is the actual MTU configured on the ports. Packets exceeding this
                               // size will be dropped.
                               // Note: in Ethernet or Fibre Channel, oper_mtu is equal to admin_mtu. 
                               //       In InfiniBand the actual operational MTU is reported which may be smaller than admin_mtu
        u16 reserved4;         //
#else
        u16 reserved2;
        u8 local_port;
        u8 reserved1;
        u16 admin_mtu;
        u16 max_mtu;
        u16 oper_mtu;
        u16 reserved3;
#endif
};

/* PTYS - Port Type and Speed Register */
#define ETH_40GBASE_KR4_FEC   0x00000400
#define ETH_40GBASE_KR4       0x00000200
#define ETH_INTERNAL_KR2_FEC  0x00000100
#define ETH_INTERNAL_KR2      0x00000040
#define ETH_10GBASE_KR_FEC    0x00000040
#define ETH_10GBASE_KR        0x00000020
#define ETH_10GBASE_KX4       0x00000010
#define ETH_10GBASE_CX4       0x00000008
#define ETH_1GBASE_KX         0x00000004
#define ETH_XAUI              0x00000002
#define ETH_SGMII             0x00000001
#define FC_8GFC 0x00000008
#define FC_4GFC 0x00000004
#define FC_2GFC 0x00000002
#define FC_1GFC 0x00000001
struct reg_ptys_layout {
#ifdef __BIG_ENDIAN
        u8 reserved1;          //
        u8 local_port;         // Local port number Index
        u16 reserved2;         //
#else
        u16 reserved2;
        u8 local_port;
        u8 reserved1;
#endif
        u32 reserved3;         //
        /* Link capabilities supported. Reported as Read Only. */
        u32 eth_proto_capability; // Etherenet port speed/protocols supported (bitmask) RO
        u32 fc_proto_capability;  // FC port speed/protocols supported (bitmask) RO
        u32 ib_proto_capability;  // IB port speed/protocols supported (bitmask) <TBD> RO
        /* Link administratively enabled capabilities by software. 
           Same encoding as <XXX>_capability fields.
           HW furthers negotiates the enabled capabilities with remote end of the link */
        u32 eth_proto_admin;      // Ethernet port speed/protocols bitmask. RW
        u32 fc_proto_admin;       // FC port speed/protocols bitmask. RW
        u32 ib_proto_admin;       // IB port speed/protocols bitmask. <TBD> RW
        /* Link operational mode. Same encoding as <XXX>_capability fields. */
        u32 eth_proto_oper;       // Ethernet port speed/protocols bitmask. RO
        u32 fc_proto_oper;        // FC port speed/protocols bitmask. RO
        u32 ib_proto_oper;        // IB port speed/protocols bitmask. <TBD> RO
        u8 reserved4[64];         //
};

/* PMLP - Ports Module to Local Port Register */
#define UNMAP_LOCAL_PORT 0x00
#define USE_LANE_0       0x01
#define USE_LANE_0_1     0x01
#define USE_LANE_0_1_2_3 0x01
struct reg_pmlp_layout {
#ifdef __BIG_ENDIAN
        u8 reserved1;          //
        u8 local_port;         // Local port number Index
        u8 reserved2;          //
        u8 width;              // RW Lane width
        u16 reserved3:14;      //
        u16 lane_0:2;          // RW Lane - up to 4 serdeses in a module can be mapped to a local port.
        u8 reserved4; 
        u8 module_0;           // Module number RW
        u16 reserved5:14;      //
        u16 lane_1:2;          // RW Lane - up to 4 serdeses in a module can be mapped to a local port.
        u8 reserved6; 
        u8 module_1;           // Module number RW
        u16 reserved7:14;      //
        u16 lane_2:2;          // RW Lane - up to 4 serdeses in a module can be mapped to a local port.
        u8 reserved8; 
        u8 module_2;           // Module number RW
        u16 reserved9:14;      //
        u16 lane_3:2;          // RW Lane - up to 4 serdeses in a module can be mapped to a local port.
        u8 reserved10; 
        u8 module_3;           // Module number RW
#else
        u8 width;
        u8 reserved2;
        u8 local_port;
        u8 reserved1;
        u8 module_0;
        u8 reserved4;
        u16 lane_0:2;
        u16 reserved3:14;
        u8 module_1;
        u8 reserved6;
        u16 lane_1:2;
        u16 reserved5:14;
        u8 module_2;
        u8 reserved8;
        u16 lane_2:2;
        u16 reserved7:14;
        u8 module_3;
        u8 reserved10;
        u16 lane_3:2;
        u16 reserved9:14;
#endif
        u16 reserved11[44];
};

/* PLIB - Port Local Port to IB Port Mapping Register */
#define PORT_UNMAPPING 0x0 // Unmap local_port from ib_port
#define PORT_MAPPING   0x1 // Create Mapping from local_port to ib_port
struct reg_plib_layout {
#ifdef __BIG_ENDIAN
        u8 map:1;          // W Map
        u8 reserved1:7;    //
        u8 local_port;     // Index, Local port number 
        u8 reserved2;      //
        u8 ib_port;        // RW IB Port remapping for local_port.
                           // FFh should be used to electronically key off an IB port.
#else
        u8 ib_port;
        u8 reserved2;
        u8 local_port;
        u8 reserved1:7;
        u8 map:1;
#endif
        u32 reserved3[3];
};

/* PAOS - Ports Administrative and Operatoinal Status Register */
#define PORT_ADMIN_STATUS_UP           0x01
#define PORT_ADMIN_STATUS_DOWN_CONFIG  0x02  // Default
#define PORT_ADMIN_STATUS_DOWN_FAILURE 0x04  // Transitioned by the hardware                     
#define PORT_OPER_STATUS_UP            0x01
#define PORT_OPER_STATUS_DOWN_CONFIG   0x02 
#define GENERATE_NOT_EVENT   0x00 // Do not generate event
#define GENERATE_UNACK_EVENT 0x01 // Generate Unacknowledged Event
#define GENERATE_ACK_EVENT   0x02 // Generate Acknowledged Event
struct reg_paos_layout {
#ifdef __BIG_ENDIAN
        u8 reserved1;
        u8 local_port;     // Index Local port number for Ethernet port
                           // InfiniBand port number for InfiniBand port
        u8 reserved2:4;
        u8 admin_status:4; // R/W   Port administrative state (the desired state of the interface)
        u8 reserved3:4;                      
        u8 oper_status:4;  // RO Port operational state
        u32 reserved4:30;
        u32 event_gen:2;   // RW Event Generation on operational state change
#else
        u8 oper_status:4;
        u8 reserved3:4;
        u8 admin_status:4;
        u8 reserved2:4;
        u8 local_port;
        u8 reserved1;
        u32 event_gen:2;
        u32 reserved4:30;
#endif
        u32 reserved5;
};

/* SPZR - Switch Port Zero Register */
#define ENCHANCED_SW_PORT_0_IGNORED  0x0
#define ENCHANCED_SW_PORT_0_VALID    0x1
#define IS_SMa                                 0x00000002
#define IS_NoticeSupported                     0x00000004
#define IS_TrapSupported                       0x00000008
#define IS_AutomaticMigrationSupported         0x00000020
#define IS_SLMappingSupported                  0x00000040
#define IS_MKeyNVRAM                           0x00000080
#define IS_PKeyNVRAM                           0x00000100
#define IS_LEDInfoSupported                    0x00000200
#define IS_SMdisabled                          0x00000400
#define IS_SystemImageGUIDSupported            0x00000800
#define IS_PKeySwitchExternalPortTrapSupported 0x00001000
#define IS_CommunicationManagementSupported    0x00010000
#define IS_SNMPTunnelingSupported              0x00020000
#define IS_ReinitSupported                     0x00040000
#define IS_DeviceManagementSupported           0x00080000
#define IS_VendorClassSupported                0x00100000
#define IS_DRNoticeSupported                   0x00200000
#define IS_CapabilityMaskNoticeSupported       0x00400000
#define IS_BootManagementSupported             0x00800000
#define IS_LinkRoundTripLatencySupported       0x01000000
#define IS_ClientReregistrationSupported       0x02000000
#define IS_OtherLocalChangesNoticeSupported    0x04000000
#define IS_LinkSpeedWidthPairsTableSupported   0x08000000 
#define SPEED_SDR                              0x01
#define SPEED_DDR                              0x02
#define SPEED_QDR                              0x04
#define SPEED_FDR10                            0x08
#define SPEED_FDR                              0x10
struct reg_spzr_layout {
#ifdef __BIG_ENDIAN
        u8 swid;             // Switch Partition to which the command applies to.
        u8 reserved1;        //
        u8 ndm:1;            // Node description mask. Set to 1 to write the NodeDescription field (see below)
        u8 enh_sw_p0_mask:1; // Enhanced Switch Port 0 mask. When 1, the EnhSwP0 field is valid. When 0,
                             // the EnhSwP0 field is ignored.
        u8 cap_mask:1;       // CM Set PortInfo:CapabilityMask to PortInfo:CapabilityMask specified.
        u8 v_key:1;          // VK Set the internal GSA V_Key. Incoming VendorSpecific MADs must have matching
                             // V_Key to the one set by this command.
        u8 mpkey:1;          // mp Change PKey table size to max_pkey
                             // This one is derived from the profile - no set.
        u8 sys_guid:1;       // sig Set System Image GUID to system_image_guid specified.
                             // System_image_guid and sig must be the same for all ports.
        u8 node_guid:1;      // ng Set node GUID to node_guid specified.
                             // node_guid and ng must be the same for all ports.
        u8 guid0:1;          // g0 Set port GUID0 to GUID0 specified
        u8 reserved2:6;         //
        u8 enh_sw_p0:1;      // EnhSwP0 When set, it enables Enhanced Switch Port 0. Reported in NodeInfo.
                             // Otherwise, Enhanced Switch Port 0 is disabled.
        u8 reserved3:1;      //
#else
        u8 reserved3:1;
        u8 enh_sw_p0:1;
        u8 reserved2:6;
        u8 guid0:1;
        u8 node_guid:1;
        u8 sys_guid:1;
        u8 mpkey:1;
        u8 v_key:1;
        u8 cap_mask:1;
        u8 enh_sw_p0_mask:1;
        u8 ndm:1;
        u8 reserved1;
        u8 swid;
#endif
        u32 capability_mask; // Sets the PortInfoCapabilityMask: Specifies the supported capabilities of this node.
                             // A bit set to 1 for affirmation of supported capability.
        u64 system_image_guid_h; // System Image GUID, takes effect only if the sig bit is set. 
                             // Must be the same for both ports.
        u64 system_image_guid_l; // System Image GUID, takes effect only if the sig bit is set.
                             // Must be the same for both ports.
        u64 guid0_h;         // EUI-64 GUID assigned by the manufacturer, takes effect only if the g0 bit is set
        u64 guid0_l;         // EUI-64 GUID assigned by the manufacturer, takes effect only if the g0 bit is set
        u64 node_guid_h;     // Node GUID, takes effect only if the ng bit is set
                             // Must be the same for both ports.
        u64 node_guid_l;     // Node GUID, takes effect only if the ng bit is set
                             // Must be the same for both ports.
        u32 v_key_h;         // V_Key_h The internal GSA V_Key. Incoming VendorSpecific MADs must have a V_Key
                             // matching the one set by this command. If not specified, then must be set to 0.
        u32 v_key_l;         // V_Key_l Takes effect only if the VK bit is set.
#ifdef __BIG_ENDIAN
        u16 max_pkey;        // max_pkey is derived from the profile - no set. Maximum pkeys for the port.
                             // Must be the same for both ports. Takes effect if the mp bit is set
        u16 reserved4;       //
        u64 node_desc;       // NodeDescription Text string that describes the node
        u8 link_speed_supported; // Supported by PTYS. Valid when lss is set.
                             // Set the PortInfo:link_speed_supported field reported by the device to the subnet
                             // manager. Set to 0 to restore factory default value.
                             // Bit 0 = SDR Bit 1 = DDR Bit 2 = QDR Bit 3 = FDR10 Bit 4 = FDR
        u8 link_width_supported:4; // Supported by PTYS/ Valid when lws is set.
                             // Set the PortInfo:link_width_supported field reported by the device to the subnet
                             // manager. Set to 0 to restore factory default value.
                             // Bit 0 = 1X Bit 1 = 4X
        u8 reserved5:4;      //
        u16 reserved6;
#else
        u16 reserved4;
        u16 max_pkey;
        u64 node_desc; // ?????
        u16 reserved6;
        u8 reserved5:4;
        u8 link_width_supported:4;
        u8 link_speed_supported;
#endif
};

/* HCAP - Host Interface Capabilities Register Register Layout */
struct reg_hcap_layout {
        u64 reserved0;
#ifdef __BIG_ENDIAN
        u16 reserved1;
        u8 reserved2;
        u8 max_num_cpu_tclass;  // Maximum number of CPU Traffic classes supported
        u16 reserved3;
        u8 reserved4;
        u8 max_num_trap_groups; // Maximum number of Trap Groups supported
        u16 reserved5;
        u8 reserved6;
        u8 reserved7:4;
        u8 max_num_dr_paths:4;  // Maximum number of Direct Route paths in the DR Path table
#else
        u8 max_num_cpu_tclass;
        u8 reserved2;
        u16 reserved1;
        u8 max_num_trap_groups;
        u8 reserved4;
        u16 reserved3;
        u8 max_num_dr_paths:4;
        u8 reserved7:4;
        u8 reserved6;
        u16 reserved5;
#endif
};

struct local_path_layout {
#ifdef __BIG_ENDIAN
        u8 reserved1;
        u8 reserved2:4;
        u8 cpu_tclass:4;
        u8 reserved3:4;
        u8 reserved4:4;
        u8 rdq:4;
#else
        u8 rdq:4;
        u8 reserved4:4;
        u8 reserved3:4;
        u8 cpu_tclass:4;
        u8 reserved2:4;
        u8 reserved1;
#endif
};

struct stacking_en_path_layout {
#ifdef __BIG_ENDIAN
        u8 reserved1:5;
        u8 stacking_tclass:3;
        u8 reserved2:4;
        u8 cpu_tclass:4;
        u8 reserved3;
        u8 reserved4:3;
        u8 rdq:5;
        u16 reserved5;
        u16 cpu_ssystem_port;
#else
        u8 rdq:5;
        u8 reserved4:3;
        u8 reserved3;
        u8 cpu_tclass:4;
        u8 reserved2:4;
        u8 reserved1:5;
        u8 stacking_tclass:3;
        u16 cpu_ssystem_port;
        u16 reserved5;
#endif
};

struct direct_route_layout {
#ifdef __BIG_ENDIAN
        u16 reserved1;
        u8 reserved2;
        u8 reserved3:4;
        u8 dr_ptr:4;  // A pointer to the Direct Route Path Table - HDRT - Host Direct Route Table
#else
        u8 dr_ptr:4;
        u8 reserved3:4;
        u8 reserved2;
        u16 reserved1;
#endif
};

struct ethernet_path_layout {
#ifdef __BIG_ENDIAN
        u16 reserved1;
        u8 mac0;
        u8 mac1;
        u8 mac2;
        u8 mac3;
        u8 mac4;
        u8 mac5;
        u16 reserved2;
        u16 reserved3:4;
        u16 vid:12;
#else
        u8 mac1;
        u8 mac0;
        u16 reserved1;
        u8 mac5;
        u8 mac4;
        u8 mac3;
        u8 mac2;
        u16 vid:12;
        u16 reserved3:4;
        u16 reserved2;
#endif
};

/* HTGT - Host Trap Group Table Register Layout */
#define CPU_PATH_LOCAL        0x00 // Used for locally attached CPU.
                                   // Mainly intended for InfinBand devices that are locally managed.
#define CPU_PATH_STACKING_EN  0x01 // Describes a path throught an Ethernet stacking domain.
                                   // Used for Ethernet domains.
#define CPU_PATH_DIRECT_ROUTE 0x02 // Describes a hop by hop route to the designated CPU. Can traverse any fabric.
#define CPU_PATH_ETHERNET     0x03 // Describes a {MAC,VLAN} path to CPU over a standard Ethernet network.
#define CPU_PATH_INTERRUPT    0x07 // Use hardware interrupt.
struct reg_htgt_layout {
#ifdef __BIG_ENDIAN
        u8 swid;        // Switch Partition ID. Stacking ports are using swid 254.
                        // To recieve events associated with stacking ports, software should configure
                        // swid 254 trap groups.
        u8 reserved0;
        u8 reserved1:4;
        u8 type:4;      // CPU Path Type
        u8 reserved2:4;
        u8 grp:4;       // Trap Group
        u16 reserved3;
        u8 pide:1;      // Policer Enable
        u8 reserved4:7;
        u8 pid;         // Global Policer ID for Trap Group grp
#else
        u8 grp:4;
        u8 reserved2:4;
        u8 type:4;
        u8 reserved1:4;
        u8 reserved0;
        u8 swid;
        u8 pid;
        u8 reserved4:7;
        u8 pide:1;
        u16 reserved3;
#endif
        u64 reserved;
        union {     // Path to designated CPU:
            struct local_path_layout local_path;
            struct stacking_en_path_layout en_path;
            struct direct_route_layout direct_route;
            struct ethernet_path_layout ethernet;
            u64 path[13];
        } path;
};

/* HPKT - Host PacKet Trap Register Layout */
#define TRAP_ACT_IGNORE  0x00 // ignore
#define TRAP_ACT_CPU     0x01 // trap to cpu (cpu receives the sole copy)
#define TRAP_ACT_MIR_CPU 0x02 // mirror to cpu (cpu receives a replica of the packet)
#define TRAP_ACT_DISCRAD 0x03 // discard
struct reg_hpkt_layout {
#ifdef __BIG_ENDIAN
        u8 reserved0:7;
        u8 ack:1;         // Ack required. Only relevant for EVENT trap_id. Indicates the device
                          // will wait for the event to be acked by the reciever.
        u8 reserved1:1;
        u8 action:3;      // Action to perform on trap_id. This field is not relevant for EVENT trap_id
                          // and must be set to 0.
        u8 reserved2:4;
        u16 trap_group:4; // Trap Group configured for trap_id
        u16 reserved3:3;
        u16 trap_id:9;    // Trap ID to configure
#else
        u16 trap_id:9;
        u16 reserved3:3;
        u16 trap_group:4;
        u8 reserved2:4;
        u8 action:3;
        u8 reserved1:1;
        u8 ack:1;
        u8 reserved0:7;
#endif
        u32 reseved[3];
};

/* HDRT - Host Direct Route Table Register Layout */
struct reg_hdrt_layout {
#ifdef __BIG_ENDIAN
        u8 reserved0:4;
        u8 dr_index:4;     // Index into the Direct Route table
        u8 reserved1;
        u8 reserved2:1;
        u8 hop_cnt:7;      // Hop Count is used to contain the number of valid elements in path and rpath
        u8 reserved3;
#define CONV_DEST_PORT(ind) ( ind )
        u8 path[64];       // Destination port number on forward path (see dir)
        u8 rpath[64];      // Destination port number on reverse path (see dir)
#else
        u8 reserved3;
        u8 hop_cnt:7;
        u8 reserved2:1;
        u8 reserved1;
        u8 dr_index:4;
        u8 reserved0:4;
#define CONV_DEST_PORT(ind) ( (ind & 0x3c) + ((ind & 0x3) ^ 0x03) )
        u8 path[64];
        u8 rpath[64];
#endif
};

/* MFCR - Misc Fan Control Register Layout */
#define PWR_FREQ_11   0x00 // b0000000 11Hz
#define PWR_FREQ_14_7 0x01 // b0000001 14.7Hz
#define PWR_FREQ_22_1 0x02 // b0000010 22.1Hz
#define PWR_FREQ_1_4  0x40 // b1000000  1.4KHz
#define PWR_FREQ_5    0x41 // b1000001  5KHz
#define PWR_FREQ_20   0x42 // b1000010 20KHz
#define PWR_FREQ_22_5 0x43 // b1000011 22.5KHz
#define PWR_FREQ_25   0x44 // b1000100 25KHz
struct reg_mfcr_layout {
#ifdef __BIG_ENDIAN
        u32 reserved0:25;
        u32 pwm_frequency:7; // Controls the frequency of the PWM signal. R/W
        u16 reserved1:6;
        u16 tacho_active:10; // Indicates which of the tachometer is active (bit per tachometer). RO
        u16 reserved2:11;
        u16 pwm_active:5;    // Indicates which of the PWM control is active (bit per PWM). RO
#else
        u32 pwm_frequency:7;
        u32 reserved0:25;
        u16 pwm_active:5;
        u16 reserved2:11;
        u16 tacho_active:10;
        u16 reserved1:6;
#endif
};

/* MFSC - Misc Fan Speed Control Register Layout */
struct reg_mfsc_layout {
#ifdef __BIG_ENDIAN
        u8 reserved0:5;
        u8 fan:3;          // Fan to control / monitor
        u8 reserved1;
        u16 reserved2;
        u16 reserved3;
        u8 reserved4;
        u8 pwm_duty_cycle; // Controls the duty cycle of the PWM. Value range from 0..255 to 
                           // represent duty cycle of 0%..100%. RW
#else
        u16 reserved2;
        u8 reserved1;
        u8 fan:3;
        u8 reserved0:5;
        u8 pwm_duty_cycle;
        u8 reserved4;
        u16 reserved3;
#endif
};

/* Misc Fan Speed Measurement Register Layout */
struct reg_mfsm_layout {
#ifdef __BIG_ENDIAN
        u8 reserved0:4;
        u8 tacho:4;        // Fan tachometer index Index
        u8 reserved1;
        u8 reserved2;
        u8 reserved3:6;
        u8 n:2;            // Defines the number of tacho pulses duration in which the SwitchX counts
                           // TBD mSec perriods. R/W
        u16 reserved4;
        u16 rpm;           // Fan speed (Round Per Minute) calculated based on the time measurement between n
                           // fan pulses. Note that in order for the RPM to be correct, the n value should
                           // correspond to the number of Tach pulses per rotation measured by the tachometer. RO
#else
        u8 num:2;
        u8 reserved3:6;
        u8 reserved2;
        u8 reserved1;
        u8 tacho:3;
        u8 reserved0:5;
        u16 rpm;
        u16 reserved4;
#endif
};

/* MFSL - Misc Fan Speed Limit Register Layout */
#define NON_GEN_EVENT    0x00 // Do not generate event
#define GEN_EVENT        0x01 // Generate Event
#define GEN_SINGLE_EVENT 0x02 // Generate single event
struct reg_mfsl_layout {
#ifdef __BIG_ENDIAN
        u8 reserved0:5;
        u8 tacho:3;        // Fan tachometer index Index
        u8 reserved1;
        u8 reserved2;
        u8 reserved3:4;
        u8 ee:2;           // Event Enable. R/W
        u8 reserved4:1;
        u8 ie:1;           // Interrupt Enable. If the ie bit is set and the Tachometer reading is out of
                           // range as defined by tach_min and tach_max, the FERR# will be asserted. R/W
        u16 reserved5;
        u16 tach_min;      // Tachometer minimum value (minimum RPM). R/W
        u16 reserved6;
        u16 tach_max;      // Tachometer maximum value (maximum RPM). R/W
#else
        u8 ie:1;
        u8 reserved4:1;
        u8 ee:2;
        u8 reserved3:4;
        u8 reserved2;
        u8 reserved1;
        u8 tacho:3;
        u8 reserved0:5;
        u16 tach_min;
        u16 reserved5;
        u16 tach_max;
        u16 reserved6;
#endif
};

/* FORE - Fan Out of Range Event  Register Layout */
struct reg_fore_layout {
#ifdef __BIG_ENDIAN
        u16 reserved0:6;
        u16 fan_under_limit:10; // Fan speed is below the low limit defined in MFSL register.
                                // Each bit relates to a single tachometer and indicated the specific
                                // tachometer reading is below the threshold. RO
        u16 reserved1;
        u16 reserved2:6;
        u16 fan_over_limit:10;  // Fan speed is above the high limit defined in MFSL register.
                                // Each bit relates to a single tachometer and indicated the
                                // specific tachometer reading is above the threshold. RO
        u16 reserved3;
#else
        u16 reserved1; 
        u16 fan_under_limit:10;
        u16 reserved0:6;
        u16 reserved3;
        u16 fan_over_limit:10;
        u16 reserved2:6;
#endif
};

/* MCIA - Misc Cable Info Access Register Layout */
struct reg_mcia_layout {
#ifdef __BIG_ENDIAN
        u8 lock:1;               // Lock bit. Setting this bit will lock the access to the specific cable.
                                 // Used for updating a full page in a cable EPROM. Any access other then
                                 // subsequence writes will fail while the port is locked. R/W
        u8 reserved0:7;
        u8 local_port;           // Local port number.
        u16 reserved1;
        u8 i2c_device_address;   // I2C device address. R/W
        u8 page_number;          // Page number. R/W
        u16 device_address;      // Device address. R/W
        u16 reserved2;
        u16 size;                // Number of bytes to read/write (up to 48 bytes). R/W
#else
        u16 reserved1;
        u8 local_port;
        u8 reserved0:7;
        u8 lock:1;
        u16 device_address;
        u8 page_number;
        u8 i2c_device_address;
        u16 size;
        u16 reserved2;
#endif
        u32 reserved3;
        u32 dwords[12];          // Bytes to read/write. RW
};

/* MTMP - Misc Temperature Register Fields */
struct reg_mtmp_layout {
#ifdef __BIG_ENDIAN
        u8 resrved0[3]; 
        u8 reserved1:4;          
        u8 sensor_index:4;       // Sensors index to access Index
        u16 reserved2;
        u16 temperature;         // Temperature reading from the sensor. Reading in 0.125 Celsius degrees. RO
        u8 mte:1;                // Max Temperature Enable - enables measuring the max temperature on a sensor RW
        u8 mtr:1;                // Max Temperature Reset - clears the value of the max temperature register RW
        u8 reserved3:6;
        u8 reserved4;
        u16 max_temperature;     // The highest measured temperature from the sensor. RO
                       // 0x00 Do not generate event NON_GEN_EVENT
                       // 0x01 Generate Event GEN_EVENT
                       // 0x02 Generate single event GEN_SINGLE_EVENT
        u8 tee:2;                // Temperature Event Enable RW
        u8 reserved5:6;
        u8 reserved6;
        u16 temperature_threshold; // If the sensor temperature measurement is above the threshold (and events are enabled),
                                   // an event will be generated. RW
#else
        u8 sensor_index:4;
        u8 reserved1:4;
        u8 resrved0[3];
        u16 temperature;
        u16 reserved2;
        u16 max_temperature;
        u8 reserved4;
        u8 reserved3:6;
        u8 mtr:1;
        u8 mte:1;
        u16 temperature_threshold;
        u8 reserved6;
        u8 reserved5:6;
        u8 tee:2;
#endif
        u32 reserved7;
};

/* MPGR - Misc Power Governor Register Fields */
#define FSPS_DISABLE 0x0 // Disable Frequency Scaling - Force device to highest frequency supported
#define FSPS_FORCE   0x1 // Force Frequency Scaling - Force device to highest frequency supported
#define FSPS_AUTO    0x2 // Auto - Automatic Frequency Scaling (based on port's speed).     
#define FSPS_HIGH    0x0 // High Frequency
#define FSPS_LOW     0x1 // Low Frequency   
#define VSPS_DISABLE 0x0 // Disable Voltage Scaling - Force device to high voltage
#define VSPS_FORCE   0x1 // Force Voltage Scaling - Force device to low voltage
#define VSPS_AUTO    0x2 // Auto - Automatic Voltage Scaling (based on device active frequency).
#define VSPS_HIGH    0x0 // High Voltage
#define VSPS_LOW     0x1 // Low Voltage
struct reg_mpgr_layout {
#ifdef __BIG_ENDIAN
        u8 resrved0[3];
        u8 reserved1:4; 
        u8 fsps_admin:4;         // Frequency Scaling Power Save Admin state:
        u8 resrved2[3]; 
        u8 reserved3:4;
        u8 fsps_status:4;        // Actual Frequency        
        u32 reserved4[2];
        u8 resrved5[3];
        u8 reserved6:4;
        u8 vsps_admin:4;         // Voltage Scaling Power Save Admin state
        u8 resrved7[3]; 
        u8 reserved8:4;
        u8 vsps_status:4;        // Actual Voltage
#else
        u8 fsps_admin:4;
        u8 reserved1:4;
        u8 resrved0[3];
        u8 fsps_status:4;
        u8 reserved3:4;
        u8 resrved2[3];
        u32 reserved4[2];
        u8 vsps_admin:4;
        u8 reserved6:4;
        u8 resrved5[3];
        u8 vsps_status:4;
        u8 reserved8:4;
        u8 resrved7[3];
#endif
};

/* MGIR Misc General Information Register Layout */
struct hw_info_layout { // 00h - 1Ch
#ifdef __BIG_ENDIAN
        u16 device_hw_revesion;
        u16 device_id;
        u16 resrved0;
        u16 resrved1:11;
        u16 dvfs:5;
        u32 resrved2[5];
        u32 uptime;
#else
        u32 b[8];
#endif
};

struct fw_info_layout { // 20h - 5Ch
#ifdef __BIG_ENDIAN
        u8 resrved0;
        u8 major;
        u8 minor;
        u8 sub_minor;
        u32 build_id;
        u8 month;
        u8 day;
        u16 year;
        u16 resrved1;
        u16 hour;
        u32 psid[4];
        u32 ini_file_version;
        u32 extended_major;
        u32 extended_minor;
        u32 extended_sub_minor;
        u32 resrved4[4];
#else
        u32 b[16];
#endif
};

struct sw_info_layout { // 60h - 7Ch
#ifdef __BIG_ENDIAN
        u8 resrved0;
        u8 major;
        u8 minor;
        u8 sub_minor;
        u32 resrved4[7];
#else
        u32 b[8];
#endif
};

struct reg_mgir_layout {
        struct hw_info_layout hw_info; // 00h - 1Ch
        struct fw_info_layout fw_info; // 20h - 5Ch
        struct sw_info_layout sw_info; // 60h - 7Ch
};

struct reg_msci_layout {
#ifdef __BIG_ENDIAN
    u8 reserved0[3];
    u8 reserved1:4; 
    u8 index:4;
#else
    u8 index:4;
    u8 reserved1:4;
    u8 reserved0[3];
#endif
    u32 version;
};
 
/* SwitchX commands opcodes */
enum switchx_cmd_opcodes {
        SX_QUERY_FW_OP        = 0x04, // QUERY_FW command opcode
        SX_QUERY_BOARDINFO_OP = 0x06, // QUERY_BOARDINFO command opcode
        SX_CONFIG_PROFILE_OP  = 0x100,// CONFIG_PROFILE command opcode TBD
        SX_ACCESS_REG_OP      = 0x40, // ACCESS_REG command opcode
        /* Not supported by I2c */
        SX_QUERY_AQ_CAP_OP    = 0x03,    
        SX_SW2HW_EQ_OP        = 0x13, 
        SX_HW2SW_EQ_OP        = 0x14,
        SX_QUERY_EQ_OP        = 0x15,
        SX_SW2HW_CQ_OP        = 0x16,  
        SX_HW2SW_CQ_OP        = 0x17,  
        SX_QUERY_CQ_OP        = 0x18,  
        SX_2ERR_DQ_OP         = 0x1E, 
        SX_QUERY_DQ_OP        = 0x22,
        SX_MAD_IFC_OP         = 0x24,
        SX_SW2HW_DQ_OP        = 0x201,  
        SX_HW2SW_DQ_OP        = 0x202,
        SX_INIT_MAD_DEMUX_OP  = 0x203,
        SX_MAP_FA_OP_OP       = 0xFFF, /// Should be some other constant
        SX_UNMAP_FA_OP_OP     = 0xFFE,       
};

enum switchx_ret_status {
        SX_STAT_OK            = 0x00,   // Command execution succeeded.
        SX_STAT_INTERNAL_ERR  = 0x01,   // Internal error (e.g. bus error) occurred while processing command.
        SX_STAT_BAD_OP        = 0x02,   // Operation/command not supported or opcode modifier not supported.
        SX_STAT_BAD_PARAM     = 0x03,   // Parameter not supported, parameter out of range.
        SX_STAT_BAD_SYS_STATE = 0x04,   // System was not enabled or bad system state.
        SX_STAT_BAD_RESOURCE  = 0x05,   // Attempt to access reserved or unallocated resource, or resource in
                                        // inappropriate ownership.
        SX_STAT_RESOURCE_BUSY = 0x06,   // Requested resource is currently executing a command.
        SX_STAT_EXCEED_LIM    = 0x08,   // Required capability exceeds device limits.
        SX_STAT_BAD_RES_STATE = 0x09,   // Resource is not in the appropriate state or ownership.
        SX_STAT_BAD_INDEX     = 0x0A,   // Index out of range (might be beyond table size or attempt to 
                                        // access a reserved resource).
        SX_STAT_BAD_NVMEM     = 0x0B,   // checksum/CRC failed.
        SX_STAT_BAD_PKT       = 0x30,   // Bad management packet (silently discarded).
};

/* QUERY_FW
 Description
 *------*------*-----*--------*------*---------*
 |Opcode|Op Mod|Event|IN PARAM|IN Mod|OUT PARAM|
 *------*------*-----*--------*------*---------*
 |0x4   |Yes   |Yes  |N/A     |Yes   |Mail box |
 *------*------*-----*--------*------*---------*
 The QUERY_FW command retrieves information related to the firmware, command interface version and 
 the amount of resources that should be   allocated to the firmware. The returned output parameter 
 depends on the OpMod value. When executing QUERY_FW through the i2c, it is required to work with
 local mailboxes. The location of the mailboxes on the i2c address space can be retrieved by 
 issuing QUERY_FW with opcode modifier 0x1.
*/

/* Opcode Modifier */
enum query_fw_opcode_mod {
        SX_MB_OUTPUT      = 0x00, // An output mailbox is returned, struct query_fw_output_mbox_layout
        SX_IMM_OUTPUT_1   = 0x01, // The output parameter is an immediate value, struct output_param_opmode1
        SX_IMM_OUTPUT_2   = 0x02, // The output parameter is an immediate values, struct output_param_opmode2
};

/*  Input Modifier
    Input Modifier for this command is relevant only if Opcode Modifier (OpMod) is equal to 0x2. 
    Otherwise reserved.
*/
enum query_fw_input_mod {
        SX_BAR0_OFF_NONE  = 0x00, // Input modifier parameter is not relevant
        SX_BAR0_OFF_BSR   = 0x01, // Output parameter returns the offset in BAR0 where the 
                                  // Boot Syndrome Register should be read from
        SX_BAR0_OFF_FER   = 0x03  // Output parameter returns the offset in BAR0 where the
                                  // Fatal Error Register should be read from
};

struct query_fw_output_mbox_layout {
#ifdef __BIG_ENDIAN
	u16 fw_pages;               // Amount of physical memory to be allocated for firmware usage
                                    // is in 4KB pages
	u16 fw_rev_major;           // Firmware Revision - Major
	u16 fw_rev_subminor;        // Firmware Sub-minor version (Patch level)
	u16 fw_rev_minor;           // Firmware Revision - Minor
	u16 core_clk;               // Internal Clock Frequency (in MHz)
	u16 cmd_interface_rev;      // Command Interface Interpreter Revision ID
                                    // This number is bumped up every time a non-backward-compatible change 
                                    // is done for the command interface. The current cmd_interface_rev is 1.
	u32 dt:1;                   // If set, Debug Trace is supported
	u32 reserved1:31;           // 
	u8 fw_hour;                 // Firmware timestamp - hour (displayed as a hexadecimal number)
	u8 fw_minutes;              // Firmware timestamp - minutes (displayed as a hexadecimal number)
	u8 fw_seconds;              // Firmware timestamp - seconds (displayed as a hexadecimal number)
	u8 reserved2;               // 
	u16 fw_year;                // Firmware timestamp - year (displayed as a hexadecimal number; e.g. 0x2005)
	u8 fw_month;                // Firmware timestamp - month (displayed as a hexadecimal number)
	u8 fw_day;                  // Firmware timestamp - day (displayed as a hexadecimal number)
	u32 reserved3;              //  
 	u32 reserved4;              // 
 	u32 clr_int_base_offset_h;  // Bits [63:32] of the Clear Interrupt register’s offset from clr_int_bar
                                    // register in PCI address space. Points to a 32-bit register.
 	u32 clr_int_base_offset_l;  // Bits [31:0] of the Clear Interrupt register’s offset from 
                                    // clr_int_bar register in PCI address space. Points to a 32-bit register
 	u32 clr_int_bar:2;          // PCI base address register (BAR) where clr_int register is located.
                                    // 00 - BAR 0-1 (64 bit BAR)
 	u32 reserved5:30;           // 
 	u32 reserved6;              // 
 	u32 error_buf_offset_h;     // Read Only buffer for internal error reports (bits [63:32] of
                                    // offset from error_buf_bar register in PCI address space.
 	u32 error_buf_offset_l;     // Read Only buffer for internal error reports (bits [31:0]
                                    // offset from error_buf_bar register in PCI address space.
 	u32 error_buf_size;         // Internal error buffer size in DWORDs
 	u32 error_buf_bar:2;        // PCI base address register (BAR) where error 
                                    // buffer register is located. 00 - BAR 0-1 (64 bit BAR)
 	u32 reserved7:30;           // 
 	u32 doorbell_page_offset_h; // msbs of the Offset of the doorbell page
 	u32 doorbell_page_offset_l; // lsbs of the Offset of the doorbell page
 	u32 doorbell_page_bar:2;    // PCI base address register (BAR) of the doorbell
                                    // page 00 - BAR 0-1 (64 bit BAR)
 	u32 reserved8:30;           // 
#else
	u16 fw_rev_major;
	u16 fw_pages;
	u16 fw_rev_minor;
	u16 fw_rev_subminor;
	u16 cmd_interface_rev;
	u16 core_clk;
	u32 reserved1:31;
	u32 dt:1;
	u8 reserved2;
	u8 fw_seconds;
	u8 fw_minutes;
	u8 fw_hour;
	u16 fw_year;                // Firmware timestamp - year (displayed as a hexadecimal number; e.g. 0x2005)
	u8 fw_month;                // Firmware timestamp - month (displayed as a hexadecimal number)
	u8 fw_day;                  // Firmware timestamp - day (displayed as a hexadecimal number)

	u32 reserved3;
 	u32 reserved4;
 	u32 clr_int_base_offset_h;
 	u32 clr_int_base_offset_l;
 	u32 reserved5:30; 
 	u32 clr_int_bar:2;
 	u32 reserved6;
 	u32 error_buf_offset_h;
 	u32 error_buf_offset_l;
 	u32 error_buf_size;
 	u32 reserved7:30;
 	u32 error_buf_bar:2;
 	u32 doorbell_page_offset_h;
 	u32 doorbell_page_offset_l;
 	u32 reserved8:30;
 	u32 doorbell_page_bar:2;
#endif
 	u8 reserved9[180];          // 
};


struct param_h_opmode1 {
#ifdef __BIG_ENDIAN
 	u32 local_mb_size:12;        // Size in bytes of the Local Input/Output Mailbox.
                                     // Bigger mailboxes cannot use local mailboxes
 	u32 local_mb_bar0_offset:20; // Bits 19:2 of offset in Bar0 where the Local Input/Output
                                     // Mailboxes should be written into/read from.
                                     // Offset bits 1:0 and 63:20 are all 0 implicitly.
#else
 	u32 local_mb_bar0_offset:20;
 	u32 local_mb_size:12;
#endif
};

// Format for Immediate Output Parameter With OpMod=0x1
struct output_param_opmode1 {
 	struct param_h_opmode1 out_param_h; // bits 63:32 of local output mailbox - cir_reg out_param_l 
 	struct param_h_opmode1 out_param_l; // bits 31: 0 of local output mailbox - cir_reg out_param_h
};

struct param_h_opmode2 {
#ifdef __BIG_ENDIAN
 	u32 reserved1:12;  
 	u32 local_mb_bar0_offset:20; // The returned output parameter depends 
                                     // on Input Modifier:
                                     // - InMod=0x1: Bits 19:2 of offset in BAR0 where the
                                     //              Boot Syndrome Register should be read from.
                                     // - InMod=0x3: Bits 19:2 of offset in BAR0 where the
                                     //              Fatal Error Register should beread from.
                                     // For the offsets described above; offset bits 1:0 and
                                     // 63:20 are all 0 implicitly.
 	//u32 reserved2:2;
#else
 	//u32 reserved2:2;
 	u32 local_mb_bar0_offset:20;
 	u32 reserved1:12;
#endif
};

// Format for Immediate Output Parameter With OpMod=0x2
struct output_param_opmode2 {
 	u32 out_param_h;                    // bits 63:32 of output parameter (reserved)
 	struct param_h_opmode2 out_param_l; // bits 31: 0 of output parameter
};

/* QUERY_BOARDINFO
 Description
 *------*------*-----*--------*------*---------*
 |Opcode|Op Mod|Event|IN PARAM|IN Mod|OUT PARAM|
 *------*------*-----*--------*------*---------*
 |0x6   |N/A   |Y    |N/A     |N/A   |Mail box |
 *------*------*-----*----*----------*---------*
 The QUERY_BOARDINFO command retrieves adapter specific parameters. This information is
 used by the driver in order to clear interrupt signalling by the device (see clr_int -
 Clear Interrupt Register. The Clear Interrupt register is used when PCI interrupts are used.
 The register is used to tell the device to deassert its interrupt. Not relevant for I2C).
 The data returned in the output mailbox is summarized in the following table:
*/

/* The PSID field is a 16-ascii (byte) character string which acts as the board ID.
   Example: AaPSID for Mellanox’s MIS5030Q_A1 board is MT_0D00110012, where:
            MT_ is the Mellanox Vendor Symbol
            0D0 is the MT_0D00110012 Board Type Symbol
            011 is the Board Version Symbol
            0012 is the Parameter Set Number
   The byte order is as follows:
    bits 31:24 at offset 0 represent the first character (byte),
    bits 23:16 - second character, etc.
    bits 31:24 at offset 4 represent the fourth character, etc.
*/

struct psid_layout {
#ifdef __BIG_ENDIAN
        u8 vendor_symbol_0;         // Vendor Symbol (VS) - 3 characters
        u8 vendor_symbol_1;
        u8 vendor_symbol_2;
        u8 board_type_symbol_0;     // Board Type Symbol (BT) - 3 characters
        u8 board_type_symbol_1;
        u8 board_type_symbol_2;
        u8 board_version_symbol_0;  // Board Version Symbol (BV) - 3 characters
        u8 board_version_symbol_1;
        u8 board_version_symbol_2;
        u8 param_set_number_0;     // Parameter Set Number (PS) - 4 characters
        u8 param_set_number_1;
        u8 param_set_number_2;
        u8 param_set_number_3;
        u8 reserved1[3];            // Reserved (R) - 3 characters (filled with zeroes)
#else
        u8 board_type_symbol_0;
        u8 vendor_symbol_2;
        u8 vendor_symbol_1;
        u8 vendor_symbol_0;
        u8 board_version_symbol_1;
        u8 board_version_symbol_0;
        u8 board_type_symbol_2;
        u8 board_type_symbol_1;
        u8 param_set_number_2;
        u8 param_set_number_1;
        u8 param_set_number_0;
        u8 reserved1[3];
        u8 param_set_number_3;
#endif
};

struct query_boardinfo_layout {
        u8 reserved1[16];           // 
#ifdef __BIG_ENDIAN
	u8 intapin;                 // When PCIe interrupt messages are being used, this value is used for clearing
                                    // an interrupt. To clear an interrupt, the driver should write the value (1<<intpin) into the
                                    // clr_int register. When using MSI-X, this register is not used
	u8 reserved2;               // 
	u16 reserved3;              // 
        u8 reserved4[4];            // 
	u16 reserved5;              // 
	u16 vsd_vendor_id;          // PCISIG Vendor ID of the vendor specifying/formatting the VSD.
                                    // The vsd_vendor_id identifies the management domain of the vsd/psid data.
                                    // Differnet vendors may choose different vsd/psid format and encoding as long
                                    // as they use their assigned vsd_vendor_id.
                                    // The psid format as described below is used in conjunction with 
                                    // Mellanox vsd_vendor_id (15B3h).
#else
	u16 reserved3;
	u8 reserved2;
	u8 intapin;
        u8 reserved4[4];
	u16 vsd_vendor_id;
	u16 reserved5;
#endif
	u8 vsd[208];                // Vendor Specific Data. The VSD string that is burnt to the Flash with the Firmware. 
	struct psid_layout psid;    // The PSID field is a 16-ascii (byte) character string which acts as the board ID.
};


/* CONFIG_PROFILE
 Description
 *------*------*-----*--------*------*---------*
 |Opcode|Op Mod|Event|IN PARAM|IN Mod|OUT PARAM|
 *------*------*-----*--------*------*---------*
 |      |Yes   |No   |Mail box|N/A   |Mail box |
 *------*------*-----*--------*-------*--------*
 The CONFIG_PROFILE command sets and queries the Switch Profile. The set command can be
 executed on the device only once at startup in order to allocate and configure all switch resources
 and prepare it for operational mode. It is not possible to change the device profile after the chip is
 in operational mode.
 Failure of the CONFIG_PROFILE command leaves the Hardware in an indeterminate state. It is
 required to perform software reset to the device following an unsuccessful completion of the
 CONFIG_PROFILE command.
 It is required to perform software reset to the device in order to change an existing profile.
 CONFIG_PROFILE command use the following mailbox for setting or querying the switch profile.
 When using Set() the command takes the mailbox from input parameter and the output parameter is reserved. 
 When using Get() the command reports the mailbox into the output parameter and
 the input parameter is reserved
*/

/* Opcode Modifier */
enum config_profile_opcode_mod {
        SX_CONF_PROFILE_SET = 0x01, // Set profile
        SX_CONF_PROFILE_GET = 0x02, // Get profile
};

/* Config Profile Switch Partition Configuration Layout */
#define SWID_PARTITION_TYPE_DISABLED    0x0000  // Switch Partition does not exist)
#define SWID_PARTITION_TYPE_INFINIBAND  0x0001
#define SWID_PARTITION_TYPE_ETHERNET    0x0010
struct swid_layout {
#ifdef __BIG_ENDIAN
        u8 mask;                   // Modify Switch Partition Configuration mask. When set,
                                   // the configuration value for the Switch Partition are taken
                                   // from the mailbox. When clear, the current configuration values are used.
                                   // Reserved on Get() command. Bit 0 - set type Other - reserved
        u8 type:4;                 // Switch Partition type
        u8 reserved1:4;            // 
        u8 reserved2;              // 
        u8 properties_i;           // ??? check location 
#else
        u8 properties_i;
        u8 reserved2;
        u8 reserved1:4;
        u8 type:4;
        u8 mask;
#endif
        u32 reserved3;             // 
};

/* Config Profile Mailbox Layout */
#define SET_MAX_VEPA_CHANNELS 0x0000000000000001 // bit 0
#define SET_MAX_LAG           0x0000000000000010 // bit 1
#define SET_MAX_PORT_PER_LAG  0x0000000000000100 // bit 2 
#define SET_MAX_MID           0x0000000000001000 // bit 3
#define SET_MAX_PGT           0x0000000000010000 // bit 4
#define SET_MAX_SYSTEM_PORT   0x0000000000100000 // bit 5
#define SET_MAX_ACTIVE_VLANS  0x0000000001000000 // bit 6
#define SET_MAX_REGIONS       0x0000000010000000 // bit 6
#define SET_MAX_IB_MC         0x0001000000000000 // bit 12
#define SET_MAX_PKEY          0x0010000000000000 // bit 13
struct config_profile_mbox_layout {
        u32 set_mask[4];            // Capability bitmask for Set() command. Setting a bit to 1 
                                    // configures the profile according to the mailbox contents. 
                                    // Setting a bit to 0 leaves the current configuration.
                                    // This field is reserved on Get() command.
                                    // Bits 8 - 11 and others are reserved
#ifdef __BIG_ENDIAN
        u8 reserved1[3];            // 
        u8 max_vepa_channels;       // Maximum Number of VEPA Channels per port (0 through 16).
                                    // 0 Multi Channel VEPA is disabled
        u16 reserved2;              // 
        u16 max_lag;                // Maximum number of LAG IDs requested
        u16 reserved3;              // 
        u16 max_port_per_lag;       // Maximum number of ports per LAG requested
        u16 reserved4;              // 
        u16 max_mid;                // Maximum Multicast IDs
                                    // Multicast IDs are allocated from 0 to max_mid-1
        u16 reserved5;              // 
        u16 max_pgt;                // Maximum records in the Port Group Table per Switch Partition
                                    // Port Group Table indexes are from 0 to max_pgt-1
        u16 reserved6;              // 
        u16 max_system_port;        // The maximum number of system ports that can be allocated.
        u32 reserved7:20;           // 
        u32 max_active_vlans:12;    // Maximum number of active VLANs for VLAN binding
        u16 reserved8;              // 
        u16 max_regions;            // Maximum number of TCAM Regions
        u8 reserved9[16];           // 
        u16 reserved10;             // 
        u16 max_ib_mc;              // Maximum number of multicast FDB records for IB FDB (in 512 chunks)
                                    // per IB Switch Partition
        u16 reserved11;             // 
        u16 max_pkey;               // Maximum per port pkey table size (for pkey enforcement)
#else
        u8 max_vepa_channels;
        u8 reserved1[3];
        u16 max_lag;
        u16 reserved2;
        u16 max_port_per_lag;
        u16 reserved3;
        u16 max_mid;
        u16 reserved4;
        u16 max_pgt;
        u16 reserved5;
        u16 max_system_port;
        u16 reserved6;
        u32 max_active_vlans:12;
        u32 reserved7:20;
        u16 max_regions;
        u16 reserved8;
        u8 reserved9[16];
        u16 max_ib_mc;
        u16 reserved10;
        u16 max_pkey;
        u16 reserved11;
#endif
        u8 reserved12[30];          //
#define MAX_SWID_PART_NUM 8
        struct swid_layout swid_config[MAX_SWID_PART_NUM];  // Configuration for Switch Partition 0-7.
        u8 reserved13[24];          // 
};

/* ACCESS_REG
 Description
 *------*------*-----*--------*------*---------*
 |Opcode|Op Mod|Event|IN PARAM|IN Mod|OUT PARAM|
 *------*------*-----*--------*------*---------*
 |      |N/A   |Y    |Mail box|N/A   |Mail box |
 *------*------*-----*--------*-------*--------*
 The ACCESS_REG command supports accessing device registers through PCIe and i2c. This
 access is mainly used for bootstrapping. Note that only a subset of registers is supported
 as listed below
*/
#define QSPTC_REG_ID SX_QSPTC_REG // 0x4009 QSPTC - QoS Stacking Port TClass Table
#define QSTCT_REG_ID SX_QSTCT_REG // 0x400B QSTCT - QoS Switch Stacking TClass Table 
#define PMLP_REG_ID  SX_PMLP_REG  // 0x5002 PMLP - Ports Module to Local Port Register
#define PMTU_REG_ID  SX_PMTU_REG  // 0x5003 PMTU - Port MTU Register
#define PTYS_REG_ID  SX_PTYS_REG  // 0x5004 PTYS - Port Type and Speed Register
#define PAOS_REG_ID  SX_PAOS_REG  // 0x5006 PAOS - Ports Administrative and Operatoinal Status Register
#define PLIB_REG_ID  SX_PLIB_REG  // 0x500A PLIB - Port Local Port to IB Port Mapping Register
#define PSPA_REG_ID  SX_PSPA_REG  // 0x500D PSPA - Port Switch Partition Allocation 
#define PELC_REG_ID  SX_PELC_REG  // 0x500E PELC - Port Extended Link Capabilities Register
#define SPZR_REG_ID  SX_SPZR_REG  // 0x6002 SPZR - Switch Port Zero Register
#define HCAP_REG_ID  SX_HCAP_REG  // 0x7001 HCAP - Host Interface Capabilities Register
#define HTGT_REG_ID  SX_HTGT_REG  // 0x7002 HTGT - Host Trap Group Table Register
#define HPKT_REG_ID  SX_HPKT_REG  // 0x7003 HPKT - Host PacKet Trap Register
#define HDRT_REG_ID  SX_HDRT_REG  // 0x7004 HDRT - Host Direct Route Table Register
#define MFCR_REG_ID  SX_MFCR_REG  // 0x9001 MFCR - Misc Fan Control Register
#define MFSC_REG_ID  SX_MFSC_REG  // 0x9002 MFSC - Misc Fan Speed Control Register
#define MFSM_REG_ID  SX_MFSM_REG  // 0x9003 MFSM - Misc Fan Speed Measurement Register
#define MFSL_REG_ID  SX_MFSL_REG  // 0x9004 MFSL - Misc Fan Speed Limit Register
#define FORE_REG_ID  SX_FORE_REG  // 0x9007 FORE - Fan Out of Range Event  Register
#define MCIA_REG_ID  SX_MCIA_REG  // 0x9006 MCIA - Misc Cable Info Access Register
#define MTMP_REG_ID  SX_MTMP_REG  // 0x900A MTMP - Misc Temperature Register
#define MPGR_REG_ID  SX_MPGR_REG  // 0x9019 MPGR - Misc Power Governor Register Fields
#define MPAT_REG_ID  SX_MPAT_REG  // 0x901A MPAT Misc Port Analyzer Table Register
#define MPAR_REG_ID  SX_MPAR_REG  // 0x901B MPAR Misc Port Analyzer Register
#define MFM_REG_ID   SX_MFM_REG   // 0x901D MFM Misc Fabric Memory  Register
#define MHSR_REG_ID  SX_MHSR_REG  // 0x901E MHSR Misc Health State Register
#define MJTAG_REG_ID SX_MJTAG_REG // 0x901F MJTAG Misc JTAG Register
#define MGIR_REG_ID  SX_MGIR_REG  // 0x9020 MGIR Misc General Information Register
#define MPFM_REG_ID  SX_MPFM_REG  // 0x9022 MPFM Misc PCIe Failure Mode Register
#define MSCI_REG_ID  SX_MSCI_REG  // 0x902A MSCI Misc System CPLD Information Register

/* EMAD TLV Types */
enum emad_tlv_types {
        SX_END_OF_TLV        = 0x00, // "END"       len=1   End of TLV list. Must be present on all packets as the last TLV.
        SX_OPERATION_TLV     = 0x01, // "OPERATION" len=4   Operation TLV, describes the method/class and response status.
        SX_DIRECT_ROUTE_TLV  = 0x02, // "DR"        len=33  Direct Route TLV. Describes the direct route path. Must only be present
                                     //                     in direct route EMADs. Must follow OPERATION TLV if present.
        SX_REG_ACCESS_TLV    = 0x03, // "REG"       len=VAR Register value. Must be present for REG_ACCESS class.
        SX_REG_USER_DATA_TLV = 0x04, // "USERDATA"  len=VAR User data for IPC class.
};

/* EMAD Classes */
enum emad_classes {
        SX_EMAD_RESERVED_CLASS   = 0x00, // Reserved Reserved for future use N/A
        SX_EMAD_REG_ACCESS_CLASS = 0x01, // Access to SwitchX registers. register_id specifies the register to be accessed.
                                         // OPERATION, (DR), REG, END
        SX_EMAD_CPU_OPER_CLASS   = 0x02, // IPC Access to CPU OPERATION, (DR), USERDATA, END
};

struct emad_eth_header {
#ifdef __BIG_ENDIAN
        u8 dmac0;             //
        u8 dmac1;             //
        u8 dmac2;             //
        u8 dmac3;             //
        u8 dmac4;             //
        u8 dmac5;             //
        u8 smac0;             //
        u8 smac1;             //
        u8 smac2;             //
        u8 smac3;             //
        u8 smac4;             //
        u8 smac5;             //
        u16 et;               //
        u8 mlx_proto;         //
        u8 ver:4;             //
        u8 reserved1:4;       //
#else
        u8 dmac3;
        u8 dmac2;
        u8 dmac1;
        u8 dmac0;
        u8 smac1; 
        u8 smac0;
        u8 dmac5;
        u8 dmac4;
        u8 smac5;
        u8 smac4;
        u8 smac3;
        u8 smac2;
        u8 reserved1:4;
        u8 ver:4;
        u8 mlx_proto;
        u16 et; 
#endif
};

enum oper_tlv_status {
        OPER_TLV_STATUS_OK           = 0x00, // Good. Operation Performed.
        OPER_TLV_STATUS_BUSY         = 0x01, // Device is busy. Can not perform the operation at the moment, requester
                                             // should retry the operation later.
        OPER_TLV_STATUS_VER_NSUPP    = 0x02, // Version not supported.
        OPER_TLV_STATUS_UNKNWN       = 0x03, // Unknown TLV.
        OPER_TLV_STATUS_REG_NSUPP    = 0x04, // Register not supported.
        OPER_TLV_STATUS_CLASS_NSUPP  = 0x05, // Class not supported.
        OPER_TLV_STATUS_METHOD_NSUPP = 0x06, // Method not supported
        OPER_TLV_STATUS_BAD_PARAM    = 0x07, // Bad parameter (e.g. port out of range, non stacking port)
        OPER_TLV_STATUS_RSRC_NAVAIL  = 0x08, // Resource not available (e.g. attempt to write to a full FDB,
                                             // allocation failed)
        OPER_TLV_STATUS_MSG_ACK      = 0x09, // Message Receipt Acknowledgement. Will return answer later.
                                             // Requester should rearm retransmission timer.
};

enum oper_tlv_method {
        OPER_TLV_METHOD_QUERY = 0x01, // Query
        OPER_TLV_METHOD_WRITE = 0x02, // Write
        OPER_TLV_METHOD_SEND  = 0x03, // Send (response not supported)
        OPER_TLV_METHOD_EVENT = 0x05, // Event (optional response)
};

#define DIRECT_ROUTE_NOT_SET 0x0
#define DIRECT_ROUTE_SET     0x1
#define OPER_TLV_REQ         0x0
#define OPER_TLV_RESP        0x1
struct operation_tlv {
#ifdef __BIG_ENDIAN
        u16 type:5;             // Operation - from enum emad_tlv_types
        u16 len:11;             // TLV length in DWORDs
        u8 dr:1;                // Direct Route - must be set on direct route packets
        u8 status:7;            // Returned status. Must be 0 on Query/Write methods requests and Send/Traps - from
                                // enum oper_tlv_status
        u8 reserved2;           //
        u16 register_id;        // Register ID within class
        u8 res_req:1;           // Response or Request

        u8 method:7;            // Method - from enum oper_tlv_method
        u8 class;               // Class of operation - from enum emad_classes
#else
        u8 class;
        u8 method:7;
        u8 res_req:1;
        u16 register_id;
        u8 reserved2;
        u8 status:7;
        u8 dr:1;
        u16 len:11;
        u16 type:5;
#endif
        u64 tid;                // Transaction ID - 64 bit opaque value returned as is in response message
                                // This field simplifies software tracking of transactions.
};

struct end_tlv {
#ifdef __BIG_ENDIAN
        u16 type:5;             // Operation - from enum emad_tlv_types
        u16 len:11;             // Length of TLV in DWORDs
        u16 reserved1;
#else
        u16 reserved1;
        u16 len:11;
        u16 type:5;
#endif
};

/* Maximum size of swicth_reg filed within REG_ACCESS TLV */
#define SX_REG_MAX_SIZE ( MAX_EMAD_FRAME_SIZE - \
                          DIR_ROUTE_TLV_SIZE - \
                          sizeof(struct emad_eth_header) - \
                          sizeof(struct operation_tlv) - \
                          struct(struct end_tlv) \
                         ) // size in bytes
/* Maximum size of emad frame for access register EMAD frame */
#define SX_ACCESS_REG_EF ( SX_REG_MAX_SIZE + \
                           sizeof(struct operation_tlv) - \
                           sizeof(struct end_tlv) \
                         ) // size in bytes
struct reg_access_tlv {
#ifdef __BIG_ENDIAN
        u16 type:5;             // Operation - from enum emad_tlv_types
        u16 len:11;             // Length of TLV in DWORDs
        u16 reserved1;          //
#else
        u16 reserved1;
        u16 len:11;
        u16 type:5;
#endif
        void *switchx_reg;      // switchx register, referenced in register_id filed of operation TLV,
                                // variable size, maximium size iz SX_REG_MAX_SIZE
};

/* Access Register Layout */
struct access_reg_layout {
        struct operation_tlv oper_tlv;      // Operation to perform including all fields
        struct reg_access_tlv reg_accs_tlv; // Register access information
};

/* EMAD Frame Format Layout
        eth_hdr;          // Ethernet Header
        oper_tlv;         // OPERATION TLV (Mandatory)
        dr_tlv;           // DR TLV (Present in DR packets only)
        reg_access;       // REG TLV (present in REG_ACCESS class)  
        user_data;        // USERDATA TLV (present in IPC class)
        end_tlv;          // END TLV (Mandatory)
*/
struct emad_frame_access_reg_template {
        struct emad_eth_header eth_hdr;
        struct operation_tlv oper_tlv;
        struct reg_access_tlv reg_accs_tlv;
        struct end_tlv e_tlv;
};

#endif /* __SWITCHX_H_ */
