/**                                                                                                   
 *
 *  Copyright (C) 2012-2014, Mellanox Technologies Ltd.  ALL RIGHTS RESERVED.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 */
#include <linux/stddef.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/device.h>
#include <asm/uaccess.h>
#include <linux/interrupt.h>
#include <linux/irq.h>
#include <linux/of_platform.h>
#include <linux/pci_ids.h>
#include <linux/pci.h>


#include <asm/dcr.h>
#include <asm/dcr-regs.h>
#include <asm/reg.h>
#include <asm/debug.h>

#include "ppc4xx_pci.h"

/* oops_dump_reg_hook

Located in ../arch/powerpc/include/asm/system.h
Defined in as following:
extern void (*oops_dump_reg_hook)(struct pt_regs *regs, u32 mcsr, int flag);

Located in ../arch/powerpc/kernel/traps.c as empty routine
void (*oops_dump_reg_hook)(struct pt_regs *regs, u32 mcsr, int flag) = NULL;
EXPORT_SYMBOL(oops_dump_reg_hook);

Called from ../arch/powerpc/kernel/traps.c as following
int machine_check_440A(struct pt_regs *regs)
{
        unsigned long reason = get_mc_reason(regs);

        printk("Machine check in kernel mode.\n");
        if (reason & ESR_IMCP){
                printk("Instruction Synchronous Machine Check exception\n");
                mtspr(SPRN_ESR, reason & ~ESR_IMCP);

                if(oops_dump_reg_hook)
                        oops_dump_reg_hook(regs, 0, 1);
        }
        else {
                u32 mcsr = mfspr(SPRN_MCSR);
                if (mcsr & MCSR_IB)
                        printk("Instruction Read PLB Error\n");
                if (mcsr & MCSR_DRB)
                        printk("Data Read PLB Error\n");
                if (mcsr & MCSR_DWB)
                        printk("Data Write PLB Error\n");
                if (mcsr & MCSR_TLBP)
                        printk("TLB Parity Error\n");
                if (mcsr & MCSR_ICP){
                        flush_instruction_cache();
                        printk("I-Cache Parity Error\n");
                }
                if (mcsr & MCSR_DCSP)
                        printk("D-Cache Search Parity Error\n");
                if (mcsr & MCSR_DCFP)
                        printk("D-Cache Flush Parity Error\n");
                if (mcsr & MCSR_IMPE)
                        printk("Machine Check exception is imprecise\n");

                if(oops_dump_reg_hook)
                        oops_dump_reg_hook(regs, mcsr, 0);


                mtspr(SPRN_MCSR, mcsr);
        }
        return 0;
}
 
*/

#define PCI_DEVICE_ID_MELLANOX 0xc738

#define MQ0_ESH     0x0047
#define MQ0_EAUH    0x0048
#define MQ0_EALH    0x0049
#define MQ0_BAUL    0x004A
#define MQ0_CF1L    0x004B
#define MQ0_ESL     0x004C
#define MQ0_EAUL    0x004D
#define MQ0_EALL    0x004E
#define MQ0_CFBHL   0x004F
#define MQ0_BAUH    0x0050

#define I2O0_DMA_ST 0x0069 // 0x210 in PLB space
#define I2O0_ISTAT  0x006A // 0x000 in PLB space
#define I2O0_EARL
#define I2O0_EARH
#define I2O0_DMA_SEAT 0x006D // 0x270 in PLB space
#define I2O0_DMA_SEAD 0x006E
 
#define PLB4A0_ACR  0x0081
#define PLB4A0_ESRL 0x0082 
#define PLB4A0_ESRH 0x0083
#define PLB4A0_EARL 0x0084
#define PLB4A0_EARH 0x0085
#define PLB4A0_CCR  0x0088
#define PLB4A1_ACR  0x0089
#define PLB4A1_ESRL 0x008A 
#define PLB4A1_ESRH 0x008B
#define PLB4A1_EARL 0x008C
#define PLB4A1_EARH 0x008D
#define PLB4A1_CCR  0x008F

#define SDR0_CP440  0x0180

#define NDFC0_CR    0x0040
#define NDFC0_SR    0x0044

#define SSDR0_PE0_LOOP    0x0303
#define SDR0_PEn_RCSSTS   0x0305
#define SDR0_PE0_L0BISTST 0x0309

#define PEUTL_STA         0x04
#define PEUTL_PBASTA      0x08





#define SDRAM_INITPLR0  0x50    /* manual initialization control             */
#define SDRAM_INITPLR1  0x51    /* manual initialization control             */
#define SDRAM_INITPLR2  0x52    /* manual initialization control             */
#define SDRAM_INITPLR3  0x53    /* manual initialization control             */
#define SDRAM_INITPLR4  0x54    /* manual initialization control             */
#define SDRAM_INITPLR5  0x55    /* manual initialization control             */
#define SDRAM_INITPLR6  0x56    /* manual initialization control             */
#define SDRAM_INITPLR7  0x57    /* manual initialization control             */
#define SDRAM_INITPLR8  0x58    /* manual initialization control             */
#define SDRAM_INITPLR9  0x59    /* manual initialization control             */
#define SDRAM_INITPLR10 0x5a    /* manual initialization control             */
#define SDRAM_INITPLR11 0x5b    /* manual initialization control             */
#define SDRAM_INITPLR12 0x5c    /* manual initialization control             */
#define SDRAM_INITPLR13 0x5d    /* manual initialization control             */
#define SDRAM_INITPLR14 0x5e    /* manual initialization control             */
#define SDRAM_INITPLR15 0x5f    /* manual initialization control             */
#define SDRAM_RQDC      0x70    /* read DQS delay control                    */
#define SDRAM_RFDC      0x74    /* read feedback delay control               */
#define SDRAM_RDCC      0x78    /* read data capture control                 */
#define SDRAM_DLCR      0x7A    /* delay line calibration                    */
#define SDRAM_CLKTR     0x80    /* DDR clock timing                          */
#define SDRAM_WRDTR     0x81    /* write data, DQS, DM clock, timing         */
#define SDRAM_SDTR1     0x85    /* DDR SDRAM timing 1                        */
#define SDRAM_SDTR2     0x86    /* DDR SDRAM timing 2                        */
#define SDRAM_SDTR3     0x87    /* DDR SDRAM timing 3                        */
#define SDRAM_MMODE     0x88    /* memory mode                               */
#define SDRAM_MEMODE    0x89    /* memory extended mode                      */
#define SDRAM_ECCCR     0x98    /* ECC error status                          */
#define SDRAM_ECCES     SDRAM_ECCCR
#define SDRAM_CID       0xA4    /* core ID                                   */
#define SDRAM_RID       0xA8    /* revision ID                               */
#define SDRAM_FCSR      0xB0    /* feedback calibration status               */
#define SDRAM_RTSR      0xB1    /* run time status tracking                  */
#define SDRAM_MCOPT1    0x20    /* memory controller options 1               */
#define SDRAM_MCOPT2    0x21    /* memory controller options 2               */
#define SDRAM_MODT0     0x22    /* on die termination for bank 0             */
#define SDRAM_MODT1     0x23    /* on die termination for bank 1             */
#define SDRAM_MODT2     0x24    /* on die termination for bank 2             */
#define SDRAM_CODT      0x26    /* on die termination for controller         */
#define SDRAM_RTR       0x30    /* refresh timer                             */

#define DCRN_CPC0_MIRQ0  0x104 /* Master Interrupt Request 0 */
#define DCRN_CPC0_MIRQ1  0x105 /* Master Interrupt Request 1 */
#define DCRN_CPC0_MIRQ2  0x106 /* Master Interrupt Request 2 */






#define PCI_BUS(d)      (((d) >> 16) & 0xff)
#define PCI_DEV(d)      (((d) >> 11) & 0x1f)
//#define PCI_FUNC(d)     (((d) >> 8) & 0x7)
#define PCI_BDF(b,d,f)  ((b) << 16 | (d) << 11 | (f) << 8)

typedef int pci_dev_t;
static pci_dev_t bdf = 0;
static pci_dev_t devfn = 0;
static struct pci_controller *hose = NULL;
static struct ppc4xx_pciex_port *port = NULL;
static void __iomem *reg = NULL;

 
/*
21.19.7 PCI Express n Reset, Clocking, and Shutdown Status (SDR0_PEn_RCSSTS) ...........................528 OK
21.19.14 PCI Express 1 Lane n Drive Register (SDR0_PE1_LnDRV) ..........................................534 OK
21.19.26 PCI Express 1 Reset Status Register (SDR0_PE1_RSTSTA)  ........................................544 OK
21.19.31 PCI Express x Lane n Error Counter (SDR0_PEx_LnERRC) ..........................................547 OK
21.20.24 Error Status Register (PEGPLn_ESR) ............................................................556 OK
21.21.3 PLB Bus Agent Status Register (PEUTLn_PBASTA)  .................................................560 OK
21.21.17 PCI Express Port Status Register (PEUTLn_PSTA) ................................................568 OK
 
PCIX0_ERRSTS
PCIX0_PLBBESR
PCIX0_PLBBEARL
PCIX0_PLBBEARH
 
*/
/*
 * Macros for indirect DCR access
 */
#define CPR0_DCR_BASE   0x0C
#define cprcfga         (CPR0_DCR_BASE+0x0)
#define cprcfgd         (CPR0_DCR_BASE+0x1)

#define SDR_DCR_BASE    0x0E
#define sdrcfga         (SDR_DCR_BASE+0x0)
#define sdrcfgd         (SDR_DCR_BASE+0x1)

#define SDRAM_DCR_BASE  0x10
#define memcfga         (SDRAM_DCR_BASE+0x0)
#define memcfgd         (SDRAM_DCR_BASE+0x1)

#define EBC_DCR_BASE    0x12
#define ebccfga         (EBC_DCR_BASE+0x0)
#define ebccfgd         (EBC_DCR_BASE+0x1)

#define mtcpr(reg, d)   do { mtdcr(cprcfga,reg);mtdcr(cprcfgd,d); } while (0)
#define mfcpr(reg, d)   do { mtdcr(cprcfga,reg);d = mfdcr(cprcfgd); } while (0)

#define mtebc(reg, d)   do { mtdcr(ebccfga,reg);mtdcr(ebccfgd,d); } while (0)
#define mfebc(reg, d)   do { mtdcr(ebccfga,reg);d = mfdcr(ebccfgd); } while (0)

#define mtsdram(reg, d) do { mtdcr(memcfga,reg);mtdcr(memcfgd,d); } while (0)
#define mfsdram(reg, d) do { mtdcr(memcfga,reg);d = mfdcr(memcfgd); } while (0)

#define mtsdr(reg, d)   do { mtdcr(sdrcfga,reg);mtdcr(sdrcfgd,d); } while (0)
#define mfsdr(reg, d)   do { mtdcr(sdrcfga,reg);d = mfdcr(sdrcfgd); } while (0)

#ifdef CONFIG_DEBUG_SPINLOCK
DEFINE_SPINLOCK(dcr_ind_lock);
#endif  /* defined(CONFIG_PPC_DCR_NATIVE) */

static void dump_tlb_44x(void)
{
        int i;

        for (i = 0; i < PPC44x_TLB_SIZE; i++) {
                unsigned long w0,w1,w2;
                asm volatile("tlbre  %0,%1,0" : "=r" (w0) : "r" (i));
                asm volatile("tlbre  %0,%1,1" : "=r" (w1) : "r" (i));
                asm volatile("tlbre  %0,%1,2" : "=r" (w2) : "r" (i));
                printk("[%02x] %08lx %08lx %08lx ", i, w0, w1, w2);
                if (w0 & PPC44x_TLB_VALID) {
                        printk("V %08lx -> %01lx%08lx %c%c%c%c%c",
                               w0 & PPC44x_TLB_EPN_MASK,
                               w1 & PPC44x_TLB_ERPN_MASK,
                               w1 & PPC44x_TLB_RPN_MASK,
                               (w2 & PPC44x_TLB_W) ? 'W' : 'w',
                               (w2 & PPC44x_TLB_I) ? 'I' : 'i',
                               (w2 & PPC44x_TLB_M) ? 'M' : 'm',
                               (w2 & PPC44x_TLB_G) ? 'G' : 'g',
                               (w2 & PPC44x_TLB_E) ? 'E' : 'e');
                }
                printk("\n");
        }
}

static noinline void oops_dump_reg(struct pt_regs *regs, u32 mcsr, int flag)
{
        unsigned long srr0, srr1, dsisr;
        u32 rqdc, rfdc, rdcc, dlcr, clktr, wrdtr, ecces;
        u32 regs_dmp[32];
	/* mcsr value can be used like it's used in machine_check_440A,
	   in case some different input is necessary for different causes
	*/
	printk("PLB4A0_ACR=0x%08x PLB4A0_ESRL=0x%08x PLB4A0_ESRH=0x%08x PLB4A0_EARL=0x%08x PLB4A0_EARH=0x%08x PLB4A0_CCR=0x%08x\n",
		mfdcr(PLB4A0_ACR), mfdcr(PLB4A0_ESRL), mfdcr(PLB4A0_ESRH), mfdcr(PLB4A0_EARL), mfdcr(PLB4A0_EARH), mfdcr(PLB4A0_CCR));
	printk("PLB4A1_ACR=0x%08x PLB4A1_ESRL=0x%08x PLB4A1_ESRH=0x%08x PLB4A1_EARL=0x%08x PLB4A1_EARH=0x%08x PLB4A1_CCR=0x%08x\n",
		mfdcr(PLB4A1_ACR), mfdcr(PLB4A1_ESRL), mfdcr(PLB4A1_ESRH), mfdcr(PLB4A1_EARL), mfdcr(PLB4A1_EARH), mfdcr(PLB4A1_CCR));
	printk("MQ0_EAUH=0x%08x MQ0_EALH=0x%08x MQ0_BAUL=0x%08x MQ0_CF1L=0x%08x MQ0_ESL=0x%08x MQ0_EALL=0x%08x\n"
		"MQ0_EAUL=0x%08x MQ0_CFBHL=0x%08x MQ0_BAUH=0x%08x MQ0_ESH==0x%08x\n",
		mfdcr(MQ0_EAUH), mfdcr(MQ0_EALH), mfdcr(MQ0_BAUL), mfdcr(MQ0_CF1L), mfdcr(MQ0_ESL), 
		mfdcr(MQ0_EALL), mfdcr(MQ0_EAUL), mfdcr(MQ0_CFBHL), mfdcr(MQ0_BAUH), mfdcr(MQ0_ESH));
	printk("I2O0_DMA_ST=0x%08x I2O0_ISTAT=0x%08x I2O0_DMA_SEAT=0x%08x I2O0_DMA_SEAD=0x%08x\n",
		mfdcr(I2O0_DMA_ST), mfdcr(I2O0_ISTAT), mfdcr(I2O0_DMA_SEAT), mfdcr(I2O0_DMA_SEAD));
	printk("SDR0_CP440=0x%08x NDFC0_CR=0x%08x NDFC0_SR=0x%08x\n",
		mfdcr(SDR0_CP440), mfdcr(NDFC0_CR), mfdcr(NDFC0_SR));
	printk("SSDR0_PE0_LOOP=0x%08x SDR0_PEn_RCSSTS=0x%08x SDR0_PE0_L0BISTST=0x%08x\n",
		mfdcri(SDR0, SSDR0_PE0_LOOP), mfdcri(SDR0, SDR0_PEn_RCSSTS), mfdcri(SDR0, SDR0_PE0_L0BISTST));
	printk("PESDR0_440SPE_LOOP=0x%08x PESDR0_440SPE_RCSSTS=0x%08x PESDR0_440SPE_HSSL0STS=0x%08x\n"
		"PESDR0_440SPE_HSSL1STS=0x%08x PESDR1_440SPE_HSSL2STS=0x%08x PESDR0_440SPE_HSSL3STS=0x%08x\n"
		"PESDR0_440SPE_HSSL4STS=0x%08x PESDR0_440SPE_HSSL5STS=0x%08x PESDR0_440SPE_HSSL6STS=0x%08x\n", 
		mfdcri(SDR0, PESDR0_440SPE_LOOP), mfdcri(SDR0, PESDR0_440SPE_RCSSTS), mfdcri(SDR0, PESDR0_440SPE_HSSL0STS),
		mfdcri(SDR0, PESDR0_440SPE_HSSL1STS), mfdcri(SDR0, PESDR1_440SPE_HSSL2STS), mfdcri(SDR0,PESDR0_440SPE_HSSL3STS),
		mfdcri(SDR0, PESDR0_440SPE_HSSL4STS), mfdcri(SDR0, PESDR0_440SPE_HSSL5STS), mfdcri(SDR0, PESDR0_440SPE_HSSL6STS));
	printk("PESDR0_460EX_RSTSTA=0x%08x PESDR0_460EX_L0DRV=0x%08x\n"
		"PESDR1_460EX_RSTSTA=0x%08x PESDR1_460EX_L0DRV=0x%08x\n" 
		"PESDR1_460EX_L1DRV=0x%08x PESDR1_460EX_L2DRV=0x%08x PESDR1_460EX_L3DRV=0x%08x\n"
		"PESDR0_460EX_L0ERRC=0x%08x\n",
		mfdcri(SDR0, PESDR0_460EX_RSTSTA), mfdcri(SDR0, PESDR0_460EX_L0DRV), 
		mfdcri(SDR0, PESDR1_460EX_RSTSTA), mfdcri(SDR0, PESDR1_460EX_L0DRV),
		mfdcri(SDR0, PESDR1_460EX_L1DRV), mfdcri(SDR0, PESDR1_460EX_L2DRV), mfdcri(SDR0, PESDR1_460EX_L3DRV),
		mfdcri(SDR0, PESDR0_460EX_L0ERRC));
	if (port) {
		printk("PESDRn_RCSSTS=0x%08x\n",
			mfdcri(SDR0, port->sdr_base + PESDRn_RCSSTS));
		printk("DCRO_PEGPL_ESR=0x%08x\n",
			dcr_read(port->dcrs, DCRO_PEGPL_ESR)); 
		printk("PEUTL_PBCTL=0x%08x PEUTLn_PBASTA=0x%08x PEUTLn_PSTA=0x%08x\n",
			in_be32(port->utl_base + PEUTL_PBCTL), in_be32(port->utl_base + PEUTL_PBASTA),
			in_be32(port->utl_base + PEUTL_STA)/*, in_be32(port->utl_base + PEUTLn_RCSTA),
			in_be32(port->utl_base + PEUTLn_EPSTA)*/);
	}
	if (reg) {
		printk("PCIX0_ERRSTS=0x%08x PCIX0_PLBBEARL=0x%08x PCIX0_PLBBEAR=0x%08x PCIX0_STS=0x%08x\n",
			readl(reg + PCIX0_ERRSTS), readl(reg + PCIX0_PLBBEARL),
                        readl(reg + PCIX0_PLBBEARH), readl(reg + PCIX0_STS));
	}
	printk("SDR0_MIRQ0=0x%08x SDR0_MIRQ1=0x%08x SDR0_MIRQ2=0x%08x\n",
		mfdcr(DCRN_CPC0_MIRQ0), mfdcr(DCRN_CPC0_MIRQ1), mfdcr(DCRN_CPC0_MIRQ2));
	mfsdram(SDRAM_RQDC, rqdc);
	mfsdram(SDRAM_RFDC, rfdc);
	mfsdram(SDRAM_RDCC, rdcc);
	mfsdram(SDRAM_DLCR, dlcr);
	mfsdram(SDRAM_CLKTR, clktr);
	mfsdram(SDRAM_WRDTR, wrdtr);
	mfsdram(SDRAM_ECCES, ecces);
	printk("MCIF0_RQDC=0x%08x MCIF0_RFDC=0x%08x MCIF0_RDCC=0x%08x MCIF0_DLCR=0x%08x\n"
		"MCIF0_CLKTR=0x%08x, MCIF0_WRDTR=0x%08x MCIF0_ECCES=0x%08x\n",
		rqdc, rfdc, rdcc, dlcr, clktr, wrdtr, ecces);
	mfsdram(SDRAM_INITPLR0, regs_dmp[0]);
	mfsdram(SDRAM_INITPLR1, regs_dmp[1]); 
	mfsdram(SDRAM_INITPLR2, regs_dmp[2]); 
	mfsdram(SDRAM_INITPLR3, regs_dmp[3]);  
	mfsdram(SDRAM_INITPLR4, regs_dmp[4]); 
	mfsdram(SDRAM_INITPLR5, regs_dmp[5]);  
	mfsdram(SDRAM_INITPLR6, regs_dmp[6]); 
	mfsdram(SDRAM_INITPLR7, regs_dmp[7]); 
	mfsdram(SDRAM_INITPLR8, regs_dmp[8]); 
	mfsdram(SDRAM_INITPLR9, regs_dmp[9]); 
	mfsdram(SDRAM_INITPLR10, regs_dmp[10]); 
	mfsdram(SDRAM_INITPLR11, regs_dmp[11]); 
	mfsdram(SDRAM_INITPLR12, regs_dmp[12]); 
	mfsdram(SDRAM_INITPLR13, regs_dmp[13]); 
	mfsdram(SDRAM_INITPLR14, regs_dmp[14]); 
	mfsdram(SDRAM_INITPLR15, regs_dmp[15]); 
	mfsdram(SDRAM_SDTR1, regs_dmp[16]); 
	mfsdram(SDRAM_SDTR2, regs_dmp[17]); 
	mfsdram(SDRAM_SDTR3, regs_dmp[18]); 
	mfsdram(SDRAM_MMODE, regs_dmp[19]); 
	mfsdram(SDRAM_MEMODE, regs_dmp[20]); 
	mfsdram(SDRAM_CID, regs_dmp[21]);
	mfsdram(SDRAM_RID, regs_dmp[22]);
	mfsdram(SDRAM_FCSR, regs_dmp[23]);
	mfsdram(SDRAM_RTSR, regs_dmp[24]); 
        mfsdram(SDRAM_MCOPT1, regs_dmp[25]);
        mfsdram(SDRAM_MCOPT2, regs_dmp[26]);
        mfsdram(SDRAM_MODT0, regs_dmp[27]);
        mfsdram(SDRAM_MODT1, regs_dmp[28]);
        mfsdram(SDRAM_MODT2, regs_dmp[29]);
        mfsdram(SDRAM_CODT, regs_dmp[30]);
        mfsdram(SDRAM_RTR, regs_dmp[31]);
	printk("SDRAM_INITPLR0=0x%08x SDRAM_INITPLR1=0x%08x SDRAM_INITPLR2=0x%08x SDRAM_INITPLR3=0x%08x\n"
		"SDRAM_INITPLR4=0x%08x SDRAM_INITPLR5=0x%08x SDRAM_INITPLR6=0x%08x SDRAM_INITPLR7=0x%08x\n"
		"SDRAM_INITPLR8=0x%08x SDRAM_INITPLR9=0x%08x SDRAM_INITPLR10=0x%08x SDRAM_INITPLR11=0x%08x\n"
		"SDRAM_INITPLR12=0x%08x SDRAM_INITPLR13=0x%08x SDRAM_INITPLR14=0x%08x SDRAM_INITPLR15=0x%08x\n" 
		"SDRAM_SDTR1=0x%08x SDRAM_SDTR2=0x%08x SDRAM_SDTR3=0x%08x SDRAM_MMODE=0x%08x\n"
		"SDRAM_MEMODE=0x%08x SDRAM_CID=0x%08x SDRAM_RID=0x%08x SDRAM_FCSR=0x%08x\n"
		"SDRAM_RTSR=0x%08x SDRAM_MCOPT1=0x%08x SDRAM_MCOPT2=0x%08x SDRAM_MODT0=0x%08x\n"
                "SDRAM_MODT1=0x%08x SDRAM_MODT2=0x%08x SDRAM_CODT=0x%08x SDRAM_RTR=0x%08x\n",
		regs_dmp[0], regs_dmp[1], regs_dmp[2], regs_dmp[3], regs_dmp[4], regs_dmp[5], regs_dmp[6], regs_dmp[7],
		regs_dmp[8], regs_dmp[9], regs_dmp[10], regs_dmp[11], regs_dmp[12], regs_dmp[13], regs_dmp[14], regs_dmp[15],
		regs_dmp[16], regs_dmp[17], regs_dmp[18], regs_dmp[19], regs_dmp[20], regs_dmp[21], regs_dmp[22], regs_dmp[23],
		regs_dmp[24], regs_dmp[25], regs_dmp[26], regs_dmp[27], regs_dmp[28], regs_dmp[29], regs_dmp[30], regs_dmp[31]); 

        if (regs) {
                srr0 = regs->nip;
                srr1 = regs->msr;
                dsisr = mfspr(SPRN_DSISR);
                printk("SRR0  0x%016lx SRR1 0x%016lx\n", srr0, srr1);
                printk("DSISR 0x%016lx DAR  0x%016lx\n", dsisr, regs->dar);
                if(mcsr)
                        printk("MCSR=0x%08x ESR=0x%016lx\n", mcsr, mfspr(SPRN_ESR));
                if (srr1 & 0x200000)
                        printk("Signalled by SDC\n");

                if (srr1 & 0x100000) {
                        printk("Load/Store detected error:\n");
                        if (dsisr & 0x8000)
                                printk("D-cache ECC double-bit error or bus error\n");
                        if (dsisr & 0x4000)
                                printk("LSU snoop response error\n");
                        if (dsisr & 0x2000)
                                printk("MMU SLB multi-hit or invalid B field\n");
                        if (dsisr & 0x1000)
                                printk("Recoverable Duptags\n");
                        if (dsisr & 0x800)
                                printk("Recoverable D-cache parity error count overflow\n");
                        if (dsisr & 0x400)
                                printk("TLB parity error count overflow\n");
                }
        }
        if (regs)
                dump_tlb_44x();
}


static int __init oops_dump_reg_init(void)
{
        struct pci_dev *pci_dev = NULL;
        struct device_node *np = NULL;
        struct device_node *np1 = NULL;
        const u32 *pval;
#if 0
        struct resource rsrc_cfg;
        struct resource rsrc_reg;
#endif
	oops_dump_reg_hook = oops_dump_reg;
	pci_dev = pci_get_device(PCI_VENDOR_ID_MELLANOX, PCI_DEVICE_ID_MELLANOX, pci_dev);
	if (pci_dev) {
		bdf = PCI_BDF(1,0,0);
		hose = pci_dev->sysdata;
		devfn = bdf - PCI_BDF(hose->first_busno, 0, 0);
		devfn = devfn << 4;
		if (ppc4xx_pciex_ports)
			port = &ppc4xx_pciex_ports[hose->indirect_type];
	}
        for_each_compatible_node(np1, NULL, "ibm,plb-pciex") {
                pval = of_get_property(np1, "port", NULL);
                printk(KERN_INFO "PCI-X %s np=%p pval=%p\n", np1->full_name, np1, pval);
                /* Get the port number from the device-tree */
                pval = of_get_property(np1, "port", NULL);
                if (pval) {
                        printk(KERN_INFO "PCIE: port=%d\n", *pval);
                        if (*pval == 0)
                                np = np1;
                }
        }
#if 0
        oops_dump_reg(NULL, 0, 0);
#endif
#if 0
        /* Fetch config space registers address */
        if (!of_address_to_resource(np, 0, &rsrc_cfg)) {
                if (!of_address_to_resource(np, 3, &rsrc_reg)) {
                       /* Map registers */
                       reg = ioremap(rsrc_reg.start, rsrc_reg.end + 1 - rsrc_reg.start); 
                       if (reg == NULL)
                               printk("%s: Can't map registers !\n", np->full_name);                    
                }
                else {
                        printk("%s: Can't get PCI-X internal register base cfg\n", np->full_name);
                }
        }
        else {
                printk("%s:Can't get PCI-X config register base\n", np->full_name);
        }
#endif
	return 0;
}

static void __exit oops_dump_reg_exit(void)
{
	oops_dump_reg_hook = NULL;
}

module_init(oops_dump_reg_init);
module_exit(oops_dump_reg_exit);

MODULE_LICENSE("GPL");
