/**
 *
 * Copyright (C) 2012-2014, Mellanox Technologies Ltd.  ALL RIGHTS RESERVED.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 */
#include <linux/types.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/rculist.h>
#include <linux/sysfs.h>
#include <linux/kobject.h>
#include <linux/device.h>
#include "sx_glue_if.h"

/* 
   API above utilize RCU mechanism.
   RCU is a synchronization mechanism used for optimization of  read-mostly operations.
   The basic idea behind RCU is to split updates into "removal" and "reclamation" phases.
   The removal phase removes references to data items within a data structure (possibly by
   replacing them with references to new versions of these data items), and can run
   concurrently with readers. The reason that it is safe to run the removal phase concurrently
   with readers is the semantics of modern CPUs guarantee that readers will see either the old
   or the new version of the data structure rather than a partially updated reference.
   The reclamation phase does the work of reclaiming (e.g., freeing) the data items removed
   from the data structure during the removal phase. Because reclaiming data items can disrupt
   any readers concurrently referencing those data items, the reclamation phase must not start
   until readers no longer hold references to those data items. 
   Splitting the update into removal and reclamation phases permits the updater to perform the
   removal phase immediately, and to defer the reclamation phase until all readers active during
   the removal phase have completed, either by blocking until they finish or by registering a callback
   that is invoked after they finish. Only readers that are active during the removal phase need be
   considered, because any reader starting after the removal phase will be unable to gain a reference
   to the removed data items, and therefore cannot be disrupted by the reclamation phase. 
   Typical RCU update sequence are as  following: 
   1)	Remove pointers to a data structure, so that subsequent readers cannot gain a reference to it. 
   2)	Wait for all previous readers to complete their RCU read-side critical sections. 
   3)	At this point, there cannot be any readers who hold references to the data structure, so it now be
        safely be reclaimed (f.e. kfree). 
   Here are used the next APIs:
   1) rcu_assign_pointer: the updater uses this function to assign a new value to an RCU-protected pointer,
      in order to safely communicate the change in value from the updater to the reader. This function returns
      the new value, and also executes any memory-barrier instructions required for a given CPU architecture. 
   2) rcu_dereference: fetches an RCU-protected pointer, which returns a value that may then be safely dereferenced.
      It does not actually dereference the pointer, instead, it protects the pointer for later dereferencing.
      It also executes any needed memory-barrier instructions for a given CPU architecture.  
*/

void* (*sx_sgmii_register_rx_hook)(sgmii_rx rx_fn) = NULL;
int (*sx_sgmii_xmit_hook)(void* hndl, struct sk_buff *skb) = NULL;

int (*i2c_write_hook)(int i2c_dev_id, int offset, int len, u8 *in_out_buf) = NULL;
int (*i2c_read_hook)(int i2c_dev_id, int offset, int len, u8 *in_out_buf) = NULL;
int (*i2c_write_dword_hook)(int i2c_dev_id, int offset, u32 val) = NULL;
int (*i2c_read_dword_hook)(int i2c_dev_id, int offset, u32 *val) = NULL;
int (*i2c_enforce_hook)(int i2c_dev_id) = NULL;
int (*i2c_release_hook)(int i2c_dev_id) = NULL;
int (*i2c_get_local_mbox_hook)(int i2c_dev_id, u32 *mb_size_in, u32 *mb_offset_in, 
				u32 *mb_size_out, u32 *mb_offset_out) = NULL;
int (*i2c_get_fw_rev_hook)(int i2c_dev_id, u64 *fw_rev) = NULL;
int (*i2c_set_go_bit_stuck_hook)(int i2c_dev_id) = NULL;

EXPORT_SYMBOL(sx_sgmii_register_rx_hook);
EXPORT_SYMBOL(sx_sgmii_xmit_hook);
EXPORT_SYMBOL(i2c_write_hook);
EXPORT_SYMBOL(i2c_read_hook);
EXPORT_SYMBOL(i2c_write_dword_hook);
EXPORT_SYMBOL(i2c_read_dword_hook);
EXPORT_SYMBOL(i2c_enforce_hook);
EXPORT_SYMBOL(i2c_release_hook);
EXPORT_SYMBOL(i2c_get_local_mbox_hook);
EXPORT_SYMBOL(i2c_get_fw_rev_hook);
EXPORT_SYMBOL(i2c_set_go_bit_stuck_hook);

void* sx_sgmii_register_rx(sgmii_rx rx_fn)
{
	void* (*sgmii_hook)(sgmii_rx);
	sgmii_hook = rcu_dereference(sx_sgmii_register_rx_hook);
	if (sgmii_hook != NULL)
		return sgmii_hook(rx_fn);
	return NULL;
 
}
EXPORT_SYMBOL(sx_sgmii_register_rx);


int sx_sgmii_xmit(void* hndl, struct sk_buff *skb)
{
	int (*sgmii_hook)(void*, struct sk_buff*);
	sgmii_hook = rcu_dereference(sx_sgmii_xmit_hook);
	if (sgmii_hook != NULL)
		return sgmii_hook(hndl, skb);
 
	return 0;
}
EXPORT_SYMBOL(sx_sgmii_xmit);

int i2c_write(int i2c_dev_id, int offset, int len, u8 *in_out_buf)
{
	int (*wi2c_hook)(int, int, int, u8 *);
	wi2c_hook = rcu_dereference(i2c_write_hook);
	if (wi2c_hook != NULL)
		return wi2c_hook(i2c_dev_id, offset, len, in_out_buf);
	return 0;
}
EXPORT_SYMBOL(i2c_write);

int i2c_read(int i2c_dev_id, int offset, int len, u8 *in_out_buf)
{
	int (*ri2c_hook)(int, int, int, u8 *);
	ri2c_hook = rcu_dereference(i2c_read_hook);
	if (ri2c_hook != NULL)
		return ri2c_hook(i2c_dev_id, offset, len, in_out_buf);
	return 0;
}
EXPORT_SYMBOL(i2c_read);

int i2c_write_dword(int i2c_dev_id, int offset, u32 val)
{
	int (*wdwi2c_hook)(int, int, u32);
	wdwi2c_hook = rcu_dereference(i2c_write_dword_hook);
	if (wdwi2c_hook != NULL)
		return wdwi2c_hook(i2c_dev_id, offset, val);
	return 0;
}
EXPORT_SYMBOL(i2c_write_dword);

int i2c_read_dword(int i2c_dev_id, int offset, u32 *val)
{
	int (*rdwi2c_hook)(int, int, u32 *);
	rdwi2c_hook = rcu_dereference(i2c_read_dword_hook);
	if (rdwi2c_hook != NULL)
		return rdwi2c_hook(i2c_dev_id, offset, val);
	return 0;
}
EXPORT_SYMBOL(i2c_read_dword);

int i2c_enforce(int i2c_dev_id)
{
	int (*e_hook)(int);
	e_hook = rcu_dereference(i2c_enforce_hook);
	if (e_hook != NULL)
		return e_hook(i2c_dev_id);
	return 0;
}
EXPORT_SYMBOL(i2c_enforce);

int i2c_release(int i2c_dev_id)
{
	int (*r_hook)(int);
	r_hook = rcu_dereference(i2c_release_hook);
	if (r_hook != NULL)
		return r_hook(i2c_dev_id);
	return 0;
}
EXPORT_SYMBOL(i2c_release);

int i2c_get_local_mbox(int i2c_dev_id, u32 *mb_size_in, u32 *mb_offset_in, 
			u32 *mb_size_out, u32 *mb_offset_out)
{
	int (*mb_hook)(int, u32*, u32*, u32*, u32*);
	mb_hook = rcu_dereference(i2c_get_local_mbox_hook);
	if (mb_hook != NULL)
		return mb_hook(i2c_dev_id, mb_size_in, mb_offset_in, 
				mb_size_out, mb_offset_out);
	return 0;
}
EXPORT_SYMBOL(i2c_get_local_mbox);

int i2c_get_fw_rev(int i2c_dev_id, u64 *fw_rev)
{
	int (*fw_rev_hook)(int, u64 *);
	fw_rev_hook = rcu_dereference(i2c_get_fw_rev_hook);
	if (fw_rev_hook != NULL)
		return fw_rev_hook(i2c_dev_id, fw_rev);
	return 0;
}
EXPORT_SYMBOL(i2c_get_fw_rev);

int i2c_set_go_bit_stuck(int i2c_dev_id)
{
	int (*go_stuck_hook)(int);
	go_stuck_hook = rcu_dereference(i2c_set_go_bit_stuck_hook);
	if (go_stuck_hook != NULL)
		return go_stuck_hook(i2c_dev_id);
	return 0;
}
EXPORT_SYMBOL(i2c_set_go_bit_stuck);

static struct kobject *mlnx_log_kobj;
int mlnx_debug_level = 0;
EXPORT_SYMBOL_GPL(mlnx_debug_level);

static ssize_t show_mlnx_log(struct kobject *kobj, struct kobj_attribute *attr, char *buf)
{
	ssize_t rc;
	rc = sprintf(buf, "Mellanox low level log: %d\n", mlnx_debug_level);
	return rc+1;
}

static ssize_t store_mlnx_log(struct kobject *kobj, struct kobj_attribute *attr, const char *buf, size_t cnt)
{
	u32 level;
	level = (u32)simple_strtol(buf, NULL, 10);
	mlnx_debug_level = level;
	return cnt;
}

static struct kobj_attribute mlnx_log_attr = __ATTR(mlnx_log, 0644, show_mlnx_log, store_mlnx_log);
 
static int __init sx_glue_if_init(void)
{
	rcu_assign_pointer(sx_sgmii_register_rx_hook, NULL);
	rcu_assign_pointer(sx_sgmii_xmit_hook, NULL);
	rcu_assign_pointer(i2c_write_hook, NULL);
	rcu_assign_pointer(i2c_read_hook, NULL);
	rcu_assign_pointer(i2c_write_dword_hook, NULL);
	rcu_assign_pointer(i2c_read_dword_hook, NULL);
	rcu_assign_pointer(i2c_enforce_hook, NULL);
	rcu_assign_pointer(i2c_release_hook, NULL);
	rcu_assign_pointer(i2c_get_local_mbox_hook, NULL);
	rcu_assign_pointer(i2c_get_fw_rev_hook, NULL);
	rcu_assign_pointer(i2c_set_go_bit_stuck_hook, NULL);

	mlnx_log_kobj = kobject_create_and_add("mlnx_log_kobj", NULL);
	if (!mlnx_log_kobj)
		return -ENOMEM;
	if (sysfs_create_file(mlnx_log_kobj, &mlnx_log_attr.attr)) {
		printk(KERN_ERR "Failed to create sysfs group for mlnx_log\n");
		sysfs_remove_file(mlnx_log_kobj, &mlnx_log_attr.attr);
		return -EACCES;
	}

	return 0;
}

static void __exit sx_glue_if_exit(void)
{
	rcu_assign_pointer(sx_sgmii_register_rx_hook, NULL);
	rcu_assign_pointer(sx_sgmii_xmit_hook, NULL);
	rcu_assign_pointer(i2c_write_hook, NULL);
	rcu_assign_pointer(i2c_read_hook, NULL);
	rcu_assign_pointer(i2c_write_dword_hook, NULL);
	rcu_assign_pointer(i2c_read_dword_hook, NULL);
	rcu_assign_pointer(i2c_enforce_hook, NULL);
	rcu_assign_pointer(i2c_release_hook, NULL);
	rcu_assign_pointer(i2c_get_local_mbox_hook, NULL);
	rcu_assign_pointer(i2c_get_fw_rev_hook, NULL);
	rcu_assign_pointer(i2c_set_go_bit_stuck_hook, NULL);

	if (mlnx_log_kobj) {
		sysfs_remove_file(mlnx_log_kobj, &mlnx_log_attr.attr);
		kobject_del(mlnx_log_kobj);
	}
}

MODULE_AUTHOR("Vadim Pasternak <vadimp@mellanox.com>");
MODULE_DESCRIPTION("SX glue logic low level interface");
MODULE_LICENSE("GPL v2");

module_init(sx_glue_if_init);
module_exit(sx_glue_if_exit);



