#include <asm/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stropts.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <asm-generic/ioctl.h>

#include "cpld.h"
#include "cpld_ifc.h"
#include "cpld_api.h"

int cpld_gpio_tdi = CPLD_GPIO_TDI0;
int cpld_gpio_tdo = CPLD_GPIO_TDO0;

static int cpld_jtag_write_pin(int fd, int pin, int val)
{
	struct write_ioctl_param params;

	params.pin  = pin;
	params.data = val ? GPIO_OUT_1 : GPIO_OUT_0;

	return ioctl(fd, IOCTL_CPLD_GPIO_WRITE, &params);
}

static int cpld_jtag_read_pin(int fd, int pin, int dir, int *val)
{
	int rc;
	struct read_ioctl_param params;

	params.pin = pin;
	params.dir = dir;

	rc = ioctl(fd, IOCTL_CPLD_GPIO_READ, &params);
	if (rc < 0)
		return CPLD_JTAG_RC_FAIL;

	*val = params.data;

	return CPLD_JTAG_RC_OK;
}

static int cpld_jtag_set(int fd, int status)
{
    struct jtag_set_ioctl_param params;
    int rc;

    params.status = status;
    rc = ioctl(fd, IOCTL_CPLD_JTAG_SET, &params);
	if (rc < 0){
        fprintf(stderr, "Failed %s, jtag_ioctl, err: %s\n", \
                (status == CPLD_JTAG_DIS ? "disable" : "enable"), strerror(errno));
		rc = CPLD_JTAG_RC_FAIL;
    }
    else
        rc = CPLD_JTAG_RC_OK;

    return rc;
}

static inline int cpld_check_cpld_num(int cpld)
{
    if (cpld <= 0 || cpld > CPLD_MAX_NUM) {
        fprintf(stderr, "Incorrect cpld_ num %d\n", cpld);
        return CPLD_JTAG_RC_FAIL;
    }
    else
        return CPLD_JTAG_RC_OK;
}

static int cpld_select(int fd, int cpld, int status)
{
    struct cpld_select_ioctl_param  params;
    int rc;
    unsigned char reg;
    char str[10];

    memset(str, 0, sizeof(str));
    if (status == CPLD_JTAG_ENB)
        strcpy(str, "enable");
    else
        strcpy(str, "disable");

    params.cpld = (unsigned char)cpld;
    params.status = (unsigned char)status;
    rc = ioctl(fd, IOCTL_CPLD_SELECT, &params);
	if (rc < 0) {
        fprintf(stderr, "Failed %s select_ioctl cpld_num %d, err: %s\n", \
                str, cpld, strerror(errno));
		rc = CPLD_JTAG_RC_FAIL;
    }
    else
        rc = CPLD_JTAG_RC_OK;

    return rc;
}

static int cpld_create_mknod(void)
{
    struct stat buf;
    char cmd[200];
    int exit_stat, rc;

    /* Check if node is already exist */
    if (!stat(CPLD_JTAG_DEVICE, &buf))
        return CPLD_JTAG_RC_OK;

    /* Create node */
    sprintf(cmd, "/bin/mknod %s c %d 0", CPLD_JTAG_DEVICE, CPLD_JTAG_DEVICE_MAJOR);
    rc = system(cmd);
    if (rc == -1){
        fprintf(stderr, "Fail to call system command %s\n", cmd);
        return CPLD_JTAG_RC_FAIL;
    }        
    else 
    {
        exit_stat = WEXITSTATUS(rc);
        if (exit_stat != EXIT_SUCCESS) {
            fprintf(stderr, "Fail to make %s, Error %s\n", cmd, strerror(exit_stat));
            return CPLD_JTAG_RC_FAIL;
        }       
    }

    return CPLD_JTAG_RC_OK;
}

static int cpld_load_module(char* path, int jtag_enable)
{
    struct stat buf;
    char cmd[200];
    int exit_stat, rc;

    if (!path) {
        fprintf(stderr, "Null path\n");
        return CPLD_JTAG_RC_FAIL;
    }

    /* Check if module is already loaded */
    if (!stat(CPLD_JTAG_SYSFS, &buf))
        return CPLD_JTAG_RC_OK;

    /* Create node */
    if (jtag_enable)
        sprintf(cmd, "/sbin/insmod %s/%s jtag_enable=1", path, CPLD_JTAG_MODULE);
    else
        sprintf(cmd, "/sbin/insmod %s/%s", path, CPLD_JTAG_MODULE);
    rc = system(cmd);
    if (rc == -1) {
        fprintf(stderr, "Fail to call system command %s\n", cmd);
        return CPLD_JTAG_RC_FAIL;
    }        
    else 
    {
        exit_stat = WEXITSTATUS(rc);
        if (exit_stat != EXIT_SUCCESS) {
            fprintf(stderr, "Fail to make %s, Error %s\n", cmd, strerror(exit_stat));
            return CPLD_JTAG_RC_FAIL;
        }       
    }

    return CPLD_JTAG_RC_OK;
}

/* Toggle TDI
 * if a specific line is 1, the corresponding bit should be 1 else it should be zero
 *  NOTE:   tck must be changed last 
 */
int cpld_jtag_gpio_write (int fd, int tms, int tdi, int tck)
{
	int rc;
	int tck1, tck2;
	/* Read TCK */
	rc = cpld_jtag_read_pin(fd, CPLD_GPIO_TCK, GPIO_OUT, &tck1);
	if (rc < 0) {
		fprintf(stderr, "TCK read failed: %s\n", strerror(errno));
		return CPLD_JTAG_RC_FAIL;
	}
		
	/* Write TMS */
	rc = cpld_jtag_write_pin(fd, CPLD_GPIO_TMS, tms);
	if (rc < 0) {
		fprintf(stderr, "TMS write failed: %s\n", strerror(errno));
		return CPLD_JTAG_RC_FAIL;
	}

	/* Read TCK */
	rc = cpld_jtag_read_pin(fd, CPLD_GPIO_TCK, GPIO_OUT, &tck2);
	if (rc < 0) {
		fprintf(stderr, "TCK read failed: %s\n", strerror(errno));
		return CPLD_JTAG_RC_FAIL;
	}

	/* Make sure TCK value did not change */
	if (tck1 != tck2)
		printf("TCK VALUE CHANGED AFTER CHANGING TMS !!!\n");

	/* Write TDI */
	rc = cpld_jtag_write_pin(fd, cpld_gpio_tdi, tdi);
	if (rc < 0) {
		fprintf(stderr, "TDI write failed: %s\n", strerror(errno));
		return CPLD_JTAG_RC_FAIL;
	}

	/* Read TCK */
	rc = cpld_jtag_read_pin(fd, CPLD_GPIO_TCK, GPIO_OUT, &tck1);
	if (rc < 0) {
		fprintf(stderr, "TCK read failed: %s\n", strerror(errno));
		return CPLD_JTAG_RC_FAIL;
	}

	/* Make sure TCK value did not change */
	if (tck1 != tck2)
		printf("TCK VALUE CHANGED AFTER CHANGING TDI !!!\n");

	/* Write TCK */
	rc = cpld_jtag_write_pin(fd, CPLD_GPIO_TCK, tck);
	if (rc < 0) {
		fprintf(stderr, "TCK write failed: %s\n", strerror(errno));
		return CPLD_JTAG_RC_FAIL;
	}

	return CPLD_JTAG_RC_OK;
}

/* This function reads the value of the GPIO line connected to the TDO pin
 * and return its value (0 or 1) 
 */
int cpld_jtag_gpio_read(int fd)
{
	int rc;
	int val;
	
	rc = cpld_jtag_read_pin(fd, cpld_gpio_tdo, GPIO_IN, &val); 
	if (rc < 0) {
		fprintf(stderr, "TDO read failed: %s\n", strerror(errno));
		return CPLD_JTAG_RC_FAIL;
	}

	return val;
}

int cpld_jtag_io_operation(int fd, int tms, int tdi, int tck)
{
    struct jtag_io_ioctl_param params;
    int rc;

    params.tms.pin = CPLD_GPIO_TMS;
    params.tms.data = tms ? GPIO_OUT_1 : GPIO_OUT_0;
    params.tdi.pin = cpld_gpio_tdi;
    params.tdi.data = tdi ? GPIO_OUT_1 : GPIO_OUT_0;
    params.tck.pin = CPLD_GPIO_TCK;
    params.tck.data = tck ? GPIO_OUT_1 : GPIO_OUT_0;

	rc = ioctl(fd, IOCTL_CPLD_JTAG_IO, &params);
    if (rc < 0)
		return CPLD_JTAG_RC_FAIL;

	rc = params.rc;
    if (rc != CPLD_JTAG_IO_OK) {
        if (rc == CPLD_JTAG_IO_WARN) {
            fprintf(stderr, "TCK Warning during jtag io operation.\n");
            rc = CPLD_JTAG_RC_OK;
        }
        else {
            fprintf(stderr, "Error during jtag io operation.\n");
            rc = CPLD_JTAG_RC_FAIL;
        }
    }
    else
        rc = CPLD_JTAG_RC_OK;

	return rc;
}

int cpld_jtag_open(char* path, int jtag_enable, int cpld)
{
    int fd;

    if (cpld_create_mknod()) {
        fprintf(stderr, "Failed to create %s character device\n", CPLD_JTAG_DEVICE);
        return CPLD_JTAG_RC_FAIL;
    }

    if (path) {
        if (cpld_load_module(path, jtag_enable)) {
            fprintf(stderr, "Failed to load %s module\n", CPLD_JTAG_MODULE);
            return CPLD_JTAG_RC_FAIL;
        }
    }

    fd = open(CPLD_JTAG_DEVICE, O_RDWR | O_SYNC);
    if (fd  < 0) {
        fprintf(stderr, "Failed to open %s: %s\n", CPLD_JTAG_DEVICE, strerror(errno));
        return fprintf(stderr, "Failed to enable cpld_jtag\n");;
    }

    if (cpld > 0) {
        if (!cpld_check_cpld_num(cpld)) {
            if (cpld_select(fd, cpld, CPLD_JTAG_ENB)) {
                fprintf(stderr, "Failed to select cpld\n");
                close(fd);
                return CPLD_JTAG_RC_FAIL;
            }            
        }
        else {
            close(fd);
            return CPLD_JTAG_RC_FAIL;
        }
    }

    if (jtag_enable) {
        if (cpld_jtag_set(fd, CPLD_JTAG_ENB)) {
            fprintf(stderr, "Failed to enable cpld_jtag\n");
            close(fd);
            return CPLD_JTAG_RC_FAIL;
        }
    }

    return fd;
}

void cpld_jtag_close(int fd, int cpld)
{
    if (cpld > 0) {
        if (cpld_select(fd, cpld, CPLD_JTAG_DIS))
            fprintf(stderr, "Failed to deselect cpld\n");
    } else {
        if (cpld_jtag_set(fd, CPLD_JTAG_DIS))
            fprintf(stderr, "Failed to disable cpld_jtag\n");
    }

    close(fd);
}

