#include <asm/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <asm-generic/ioctl.h>

#include "cpld.h"
#include "cpld_dpt_ifc.h"
#include "cpld_api.h"

#include "sx/sxd/sxd_dpt.h"
#include <sx/sxd/kernel_user.h>
#include <sx/sxd/sxd_access_register.h>
#include <logger.h>

#define MAX_BUFFER_CNT 40
#define MAX_WRITE_CNT  40
#define MAX_READ_CNT   16

enum {
	TestLogicReset,
	RunTestIdle,
	SelectDR,
	CaptureDR,
	ShiftDR,
	Exit1DR,
	PauseDR,
	Exit2DR,
	UpdateDR,
	SelectIR,
	CaptureIR,
	ShiftIR,
	Exit1IR,
	PauseIR,
	Exit2IR,
	UpdateIR
};

const char* tap_state_names[16] = {
	"TestLogicReset",
	"RunTestIdle",
	"SelectDR",
	"CaptureDR",
	"ShiftDR",
	"Exit1DR",
	"PauseDR",
	"Exit2DR",
	"UpdateDR",
	"SelectIR",
	"CaptureIR",
	"ShiftIR",
	"Exit1IR",
	"PauseIR",
	"Exit2IR",
	"UpdateIR"
};

#define ISC_READ 0x205

int tap_state = TestLogicReset;
unsigned shift_tap_IR = 0;
unsigned tap_IR = 0; //Should be set to IDCODE by default
int tap_state_transition = 0;

struct ku_mjtag_reg mjtag_reg;
int jtag_sqn;
int last_read_tdo = 0;
int jtag_write_cnt = 0;
int jtag_read_base = 0;
int jtag_read_cnt = 0;
int jtag_read_ptr = 0;

int free_cntr = 0;

static void tap_fsm(int tms, int tdi)
{
	int next_state = tap_state;

	switch(tap_state) {
		case TestLogicReset:
			if (tms)
                next_state = TestLogicReset;
			else
                next_state = RunTestIdle;
			break;
		case RunTestIdle:
			if (tms)
                next_state = SelectDR;
			else
                next_state = RunTestIdle;
			break;
		case SelectDR:
			if (tms)
                next_state = SelectIR;
			else
                next_state = CaptureDR;
			break;
		case CaptureDR:
			if (tms)
                next_state = Exit1DR;
			else
                next_state = ShiftDR;
			break;
		case ShiftDR:
			if (tms)
                next_state = Exit1DR;
			else
                next_state = ShiftDR;
			break;
		case Exit1DR:
			if (tms)
                next_state = UpdateDR;
			else
                next_state = PauseDR;
			break;
		case PauseDR:
			if (tms)
                next_state = Exit2DR;
			else
                next_state = PauseDR;
			break;
		case Exit2DR:
			if (tms)
                next_state = UpdateDR;
			else
                next_state = ShiftDR;
			break;
		case UpdateDR:
			if (tms)
                next_state = SelectDR;
			else
                next_state = RunTestIdle;
			break;
		case SelectIR:
			if (tms)
                next_state = TestLogicReset;
			else
                next_state = CaptureIR;
			break;
		case CaptureIR:
			if (tms)
                next_state = Exit1IR;
			else
                next_state = ShiftIR;
			break;
		case ShiftIR:
			shift_tap_IR >>= 1;
			if (tdi) shift_tap_IR |= (1<<31);

			if (tms)
                next_state = Exit1IR;
			else
                next_state = ShiftIR;
			break;
		case Exit1IR:
			if (tms)
                next_state = UpdateIR;
			else
                next_state = PauseIR;
			break;
		case PauseIR:
			if (tms)
                next_state = Exit2IR;
			else
                next_state = PauseIR;
			break;
		case Exit2IR:
			if (tms)
                next_state = UpdateIR;
			else
                next_state = ShiftIR;
			break;
		case UpdateIR:
			tap_IR = (shift_tap_IR >> 22) & 0x3ff; //10 bit IR
			CPLD_LOG_DBG("Set IR = %08x\n", tap_IR);
			if (tms)
                next_state = SelectDR;
			else
                next_state = RunTestIdle;
			break;
		default:
			fprintf(stderr, "ERROR in State Transition!\n");
			break;
	}

	tap_state_transition = 0;
	if (tap_state != next_state) {
		CPLD_LOG_DBG("Transition: %s => %s\n", tap_state_names[tap_state], tap_state_names[next_state]);
		tap_state_transition = 1;
	}
	tap_state = next_state;
}

static int flush_jtag_io(int* tdo)
{
	sxd_status_t sxd_err = SXD_STATUS_SUCCESS;
	sxd_reg_meta_t reg_meta;
	int i, rc = CPLD_JTAG_RC_OK;

    reg_meta.access_cmd=SXD_ACCESS_CMD_SET;
    reg_meta.dev_id=1;

	mjtag_reg.cmd = CMD_ACCESS;
	mjtag_reg.size = jtag_write_cnt;
	mjtag_reg.seq_num = (++jtag_sqn) & 0xf;
	sxd_err = sxd_access_reg_mjtag(&mjtag_reg,&reg_meta, 1, NULL, NULL);
	if (SXD_CHECK_FAIL(sxd_err)) {
		fprintf(stderr, "JTAG CMS_ACCESS failed sxd_err=%d\n", sxd_err);
		rc = CPLD_JTAG_RC_FAIL;
	}
	*tdo = mjtag_reg.jtag_transaction_sets[jtag_write_cnt-1].tdo ? 1:0;

	CPLD_LOG_DBG("== Do Write sqn:%d cnt:%d\n",(jtag_sqn-1), jtag_write_cnt);
	for (i = 0; i < jtag_write_cnt; i++) {
		CPLD_LOG_DBG("tms,tdi,tdo: %d,%d,%d\n", mjtag_reg.jtag_transaction_sets[i].tms,\
                     mjtag_reg.jtag_transaction_sets[i].tdi, mjtag_reg.jtag_transaction_sets[i].tdo);
	}

	jtag_write_cnt = 0;
	return rc;
}

static int do_bulk_read(void)
{
	sxd_status_t sxd_err = SXD_STATUS_SUCCESS;
	sxd_reg_meta_t reg_meta;
	int i, rc = CPLD_JTAG_RC_OK;

    reg_meta.access_cmd=SXD_ACCESS_CMD_SET;
    reg_meta.dev_id=1;

	mjtag_reg.cmd = CMD_ACCESS;
	mjtag_reg.size = jtag_read_cnt + jtag_write_cnt;
	mjtag_reg.seq_num = (++jtag_sqn) & 0xf;
	sxd_err = sxd_access_reg_mjtag(&mjtag_reg,&reg_meta, 1, NULL, NULL);
	if (SXD_CHECK_FAIL(sxd_err)) {
		fprintf(stderr, "JTAG CMS_ACCESS failed sxd_err=%d\n", sxd_err);
		rc = CPLD_JTAG_RC_FAIL;
	}
	CPLD_LOG_DBG("== Do Bulk Read sqn:%d cnt:%d\n",jtag_sqn, jtag_read_cnt);
	for (i = 0; i < jtag_read_cnt + jtag_write_cnt; i++) {
		CPLD_LOG_DBG("tms,tdi,tdo: %d,%d,%d\n", mjtag_reg.jtag_transaction_sets[i].tms,\
                     mjtag_reg.jtag_transaction_sets[i].tdi, mjtag_reg.jtag_transaction_sets[i].tdo);
	}
	return rc;
}

int cpld_jtag_dpt_io_operation(int tms, int tdi, int read_tdo, int* tdo)
{
	char ch;
	sxd_status_t sxd_err = SXD_STATUS_SUCCESS;
	sxd_reg_meta_t reg_meta;
	int i, rc = CPLD_JTAG_RC_OK;

	CPLD_LOG_DBG("%07d tms,tdi,read_tdo: %d,%d,%d = \n", free_cntr++, tms,tdi,read_tdo);

	reg_meta.access_cmd=SXD_ACCESS_CMD_SET;
	reg_meta.dev_id=1;
	if (!jtag_write_cnt && !jtag_read_cnt)
		memset(&mjtag_reg, 0, sizeof(struct ku_mjtag_reg));

	if (read_tdo) {
		if (!last_read_tdo) {
			CPLD_LOG_DBG("RS rb:%02d rp:%02d rc:%02d wc:%02d\n",jtag_read_base, jtag_read_ptr,\
                         jtag_read_cnt, jtag_write_cnt);
			if ((MAX_BUFFER_CNT - jtag_write_cnt) < MAX_READ_CNT) {
                rc = flush_jtag_io(tdo);
                jtag_read_base = 0;
            } else {
                if ((ShiftDR == tap_state) && ISC_READ==tap_IR) {
                    jtag_read_base = jtag_write_cnt;
                    jtag_read_cnt = MAX_READ_CNT;
                    jtag_read_ptr = 0;

                    mjtag_reg.jtag_transaction_sets[jtag_read_base].tms = tms?1:0;
                    mjtag_reg.jtag_transaction_sets[jtag_read_base].tdi = tdi?1:0;
                    mjtag_reg.jtag_transaction_sets[jtag_read_base].tdo = 0;

                    for (i = 1; i < jtag_read_cnt; i++) {
                        mjtag_reg.jtag_transaction_sets[i+jtag_read_base].tms = 0;
                        mjtag_reg.jtag_transaction_sets[i+jtag_read_base].tdi = tdi?1:0;
                        mjtag_reg.jtag_transaction_sets[i+jtag_read_base].tdo = 0;
                    }
                    do_bulk_read();
                } else {
                    rc = flush_jtag_io(tdo);
                    jtag_read_base = 0;
                }
            }
		}
		else {
			CPLD_LOG_DBG("RK rb:%02d rp:%02d rc:%02d wc:%02d\n",jtag_read_base, jtag_read_ptr,\
                         jtag_read_cnt, jtag_write_cnt);
		}
		if (jtag_read_ptr == jtag_read_cnt) {
            jtag_read_base = 0;
			memset(&mjtag_reg, 0, sizeof(struct ku_mjtag_reg));
			if ((ShiftDR == tap_state) && ISC_READ==tap_IR) {
				jtag_read_cnt = MAX_READ_CNT;
				jtag_read_ptr = 0;
				if (tms) { //If after this operation will exit ShiftDR state
					jtag_write_cnt = 1;
					mjtag_reg.jtag_transaction_sets[0].tms = tms?1:0;
					mjtag_reg.jtag_transaction_sets[0].tdi = tdi?1:0;
				} else {
					for (i = 0; i < jtag_read_cnt; i++) {
						mjtag_reg.jtag_transaction_sets[i].tdi = tdi?1:0;
					}
				}
			} else {
				jtag_read_cnt = 1;
				jtag_read_ptr = 0;
				mjtag_reg.jtag_transaction_sets[jtag_read_ptr].tms = tms?1:0;
				mjtag_reg.jtag_transaction_sets[jtag_read_ptr].tdi = tdi?1:0;
			}
			do_bulk_read();
			*tdo = mjtag_reg.jtag_transaction_sets[jtag_read_ptr+jtag_read_base].tdo;
            jtag_read_ptr++;
		} else {
			*tdo = mjtag_reg.jtag_transaction_sets[jtag_read_ptr+jtag_read_base].tdo;
			CPLD_LOG_DBG("== Do RAM Read\n");
			CPLD_LOG_DBG("tms,tdi,tdo: %d,%d,%d\n", mjtag_reg.jtag_transaction_sets[jtag_read_ptr+jtag_read_base].tms, mjtag_reg.jtag_transaction_sets[jtag_read_ptr+jtag_read_base].tdi, mjtag_reg.jtag_transaction_sets[jtag_read_ptr+jtag_read_base].tdo);
			if ((ShiftDR == tap_state) && (ISC_READ == tap_IR) && tms) {
                CPLD_LOG_DBG("== Move operation to write on transition\n");
                jtag_read_base = 0;
                jtag_read_cnt = 0;
                jtag_read_ptr = 0;
				jtag_write_cnt = 1;
				mjtag_reg.jtag_transaction_sets[jtag_read_ptr].tms = tms?1:0;
				mjtag_reg.jtag_transaction_sets[jtag_read_ptr].tdi = tdi?1:0;
                read_tdo = 0; //Fake last operation as if it was a write operation
			}
            else {
                jtag_read_ptr++;
            }
		}
	}
	else { //!read_tdo
		if (last_read_tdo) {
			CPLD_LOG_DBG("WS rb:%02d rp:%02d rc:%02d wc:%02d\n",jtag_read_base, jtag_read_ptr,\
                         jtag_read_cnt, jtag_write_cnt);
			//flush reads
            jtag_read_base = 0;
			jtag_read_cnt = 0;
			jtag_read_ptr = 0;
			jtag_write_cnt = 0;
		}
		else {
			CPLD_LOG_DBG("WK rb:%02d rp:%02d rc:%02d wc:%02d\n",jtag_read_base, jtag_read_ptr,\
                         jtag_read_cnt, jtag_write_cnt);
		}

		mjtag_reg.jtag_transaction_sets[jtag_write_cnt].tdi = tdi?1:0;
		mjtag_reg.jtag_transaction_sets[jtag_write_cnt].tms = tms?1:0;
		jtag_write_cnt++;
		if (jtag_write_cnt == MAX_WRITE_CNT) rc = flush_jtag_io(tdo);
	}
	CPLD_LOG_DBG("-----\n");
	last_read_tdo = read_tdo;
	tap_fsm(tms, tdi);
	return rc;
}

void cpld_jtag_delay(long microseconds)
{
    sxd_status_t sxd_err = SXD_STATUS_SUCCESS;
	sxd_reg_meta_t     reg_meta;
	int tdo = 0;

	CPLD_LOG_DBG("== DELAY rb:%02d rp:%d rc:%d wc:%d\n",jtag_read_base, jtag_read_ptr,\
                 jtag_read_cnt, jtag_write_cnt);
	/* Empty buffer before JAM_DELAY */
	reg_meta.access_cmd=SXD_ACCESS_CMD_SET;
	reg_meta.dev_id=1;
	if (jtag_read_cnt) {
		jtag_read_cnt = 0;
		jtag_read_ptr = 0;
        jtag_read_base = 0;
	}
	if (jtag_write_cnt) {
		if (CPLD_JTAG_RC_OK != flush_jtag_io(&tdo)) {
			fprintf(stderr, "Failed JTAG I/O flush before JAM_DELAY\n");
		}
	}
	jtag_write_cnt = 0;
    usleep(microseconds);
}

int cpld_jtag_dpt_open(void)
{
	sxd_status_t sxd_err = SXD_STATUS_SUCCESS;
	sxd_reg_meta_t     reg_meta;

	if (lc_init()) {
		fprintf(stderr, "Failed to initialize locks\n");
		return CPLD_JTAG_RC_FAIL;
	}

    if (sxd_access_reg_init(0,log_cb , SX_LOG_INFO )) {
        fprintf(stderr, "Failed to initialize access register interface\n");
        return CPLD_JTAG_RC_FAIL;
    }

    reg_meta.access_cmd=SXD_ACCESS_CMD_SET;
    reg_meta.dev_id=1;

    memset(&mjtag_reg, 0, sizeof(struct ku_mjtag_reg));
	mjtag_reg.cmd = 1;

    sxd_err = sxd_access_reg_mjtag(&mjtag_reg,&reg_meta,1,NULL,NULL);
    if (SXD_CHECK_FAIL(sxd_err)) {
        fprintf(stderr, "Failed to enable JTAG interface %d\n", sxd_err);
        return CPLD_JTAG_RC_FAIL;
    }

	jtag_sqn = 0;
    return CPLD_JTAG_RC_OK;
}

void cpld_jtag_dpt_close(void)
{
    sxd_status_t sxd_err = SXD_STATUS_SUCCESS;
	sxd_reg_meta_t     reg_meta;
	int tdo = 0;

	if (jtag_write_cnt) {
		if (CPLD_JTAG_RC_OK != flush_jtag_io(&tdo)) 
			fprintf(stderr, "Failed JTAG I/O flush before Close.\n");
	}

	reg_meta.access_cmd = SXD_ACCESS_CMD_SET;
	reg_meta.dev_id = 1;
	memset(&mjtag_reg, 0, sizeof(struct ku_mjtag_reg));
	mjtag_reg.cmd = 0;

	sxd_err = sxd_access_reg_mjtag(&mjtag_reg,&reg_meta,1,NULL,NULL);
	if (SXD_CHECK_FAIL(sxd_err)) 
        fprintf(stderr, "Failed to disable JTAG interface %d\n", sxd_err);

	sxd_err = sxd_access_reg_deinit();
    if (SXD_CHECK_FAIL(sxd_err)) 
        fprintf(stderr, "Failed to deinit command interface, sxd_ret: %d", sxd_err);
}


